#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

##################################################################
#  WaterLevelTree Test
#  Goal: Test script for WaterLevelTree algorithm
#  Author: wenchieh
#
#  Project: eaglemine
#      waterleveltree.py
#      Version: 
#      Date: November 29 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <11/30/2017>
#
##################################################################

__author__ = 'wenchieh'


# sys
import os
import time
import argparse

# third-part lib
import numpy as np

# project
from .utils.loader import Loader
from .core.leveltree import LevelTree


VERBOSE = False
tiny_blobs = 'tiny_blob2cnt.out'
contracttree = 'level_tree_contract.out'
prunetree = 'level_tree_prune.out'
refinetree = 'level_tree_refine.out'


def waterleveltree(cel2cnt_arr, outpath):
    tsr_arr = np.asarray(cel2cnt_arr)
    values = np.log2(1.0 + tsr_arr[:, -1])
    max_level = np.max(values)
    step = 0.2

    tree = LevelTree()
    print("Construct raw-tree.")
    tree.build_level_tree(tsr_arr[:, :-1], values, 1.0, max_level, step,
                          verbose=False, outfn=os.path.join(outpath, tiny_blobs))

    print("Refine tree structure.")
    print("a). tree contract")
    tree.tree_contract(VERBOSE)
    tree.save_leveltree(os.path.join(outpath, contracttree))

    print("b). tree pruned")
    tree.tree_prune(alpha=0.8, verbose=VERBOSE)
    tree.save_leveltree(os.path.join(outpath, prunetree))

    print("c). tree node expand")
    tree.tree_node_expand(VERBOSE)
    tree.save_leveltree(os.path.join(outpath, refinetree))

    tree.dump()


if __name__ == '__main__':
    path = '../output/'
    histogram_infn = 'histogram.out'

    loader = Loader()
    print("load data")
    shape, ticks_vec, hist_arr = loader.load_multi_histogram(os.path.join(path, histogram_infn))
    mode = len(shape)
    print("Info: mode:{} shape:{}".format(mode, shape))
    waterleveltree(hist_arr, path)
    print("done!")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   _model.py
@Desc    :   Interface of class Model, DMmodel, MLmodel.
'''

# here put the import lib

class Model():
    '''Basic class of model entity.

    Attributes
    ----------
    model_name : str
        name of model
    '''

    def __init__(self, model_name: str = "my_model", *args, **kwargs):
        '''Initialization function.'''
        self.model_name = model_name

    @classmethod
    def __create__(cls, data: object, *args, **kwargs):
        '''Interface of creation by class, overrided by subclasses.

        Raises
        ----------
        NotImplementedError
            when called, return not implemented error
        '''
        raise Exception("__create__ function not implemented.")


class DMmodel(Model):
    '''Interface for data mining model.
    '''

    def __init__(self, tensor: object, *args, **kwargs):
        '''Only support construction by classmethod.
        '''
        super(DMmodel, self).__init__(*args, **kwargs)

    @classmethod
    def __create__(cls, tensor: object, *args, **kwargs) -> object:
        '''Interface of creation by class, overrided by subclasses.

        Parameters
        ----------
        tensor: object
            data object
        params: dict
            parameter dictionary

        Returns
        ----------
        _obj
            object for specific model
        '''
        _obj = cls(tensor, *args, **kwargs)
        return _obj

    def run(self, *args, **kwargs):
        '''Interface of run function, overrided by subclasses.

        Raises
        ----------
        NotImplementedError
            when called, return not implemented error
        '''
        raise Exception("Run function not implemented.")


class MLmodel(Model):
    '''Interface for machine learning model.
    '''

    def __init__(self, tensor: object, *args, **kwargs):
        '''Only support construction by classmethod.
        '''
        super(MLmodel, self).__init__(*args, **kwargs)

    @classmethod
    def __create__(cls, tensor: object, *args, **kwargs) -> object:
        '''Interface of creation by class, overrided by subclasses.

        Parameters
        ----------
        tensor: object
            data object
        params: dict
            parameter dictionary

        Returns
        ----------
        _obj
            object for specific model
        '''
        _obj = cls(tensor, *args, **kwargs)
        return _obj

    def fit(self, *args, **kwargs):
        '''Interface of fit function, overrided by subclasses.

        Raises
        ----------
        NotImplementedError
            when called, return not implemented error
        '''
        raise Exception("Fit function not implemented.")

    def predict(self, *args, **kwargs):
        '''Interface of predict function, overrided by subclasses.

        Raises
        ----------
        NotImplementedError
            when called, return not implemented error
        '''
        raise Exception("Predict function not implemented.")


class PipeLine():
    def __init__(self, data, module_list, pipe_name="my_pipeline", *args, **kwargs):
        self.module_list = module_list
        self.data = data

    def run(self):
        
        data = self.data
        for module in self.module_list:
            model, params = module
            if not isinstance(model, tuple):
                model = model(data, **params)
                if isinstance(model, DMmodel):
                    data = model.run()
                elif isinstance(model, MLmodel):
                    model.fit()
                    data = model.predict()
            else:
                task, model = model
                task = task.create(data, model, **params)
                data = task.run()
        return data
        
class Generalmodel(Model):
    '''Interface for general model.
    '''
    def __init__(self, *args, **kwargs):
        '''Only support construction by classmethod.
        '''
        super(Generalmodel, self).__init__(*args, **kwargs)

    @classmethod
    def __create__(cls, *args, **kwargs) -> object:
        '''Interface of creation by class, overrided by subclasses.

        Parameters
        ----------
        tensor: object
            data object
        params: dict
            parameter dictionary

        Returns
        ----------
        _obj
            object for specific model
        '''
        _obj = cls(*args, **kwargs)
        return _obj



import functools
import warnings

import numpy as np
import sparse
import scipy
import scipy.sparse as ssp
import torch
import torch.sparse as tsparse


def _ensure_tensor(x):
    if isinstance(x, torch.Tensor):
        if x.ndim == 0:
            return x.item()
        if x.is_sparse:
            return STensor.from_torch(x)
        return DTensor(x)
    elif isinstance(x, torch.Size):
        return DTensor(torch.as_tensor(x))
    else:
        return x


def _wrap_ret(func):
    """Wrap return value of func to spartan tensor types.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ret = func(*args, **kwargs)
        try:
            ret = torch.as_tensor(ret).cuda()
        except Exception:
            pass
        return _ensure_tensor(ret)
    return wrapper


class TorchArithmetic:
    # Unary Opeartors
    @_wrap_ret
    def __neg__(self):
        return -self._data

    @_wrap_ret
    def __pos__(self):
        return +self._data

    @_wrap_ret
    def __invert__(self):
        return ~self._data

    # Binary Operators
    @_wrap_ret
    def __add__(self, other):
        try:
            return self._data + other._data
        except AttributeError:
            return self._data + other

    @_wrap_ret
    def __mul__(self, other):
        try:
            return self._data * other._data
        except AttributeError:
            return self._data * other

    @_wrap_ret
    def __truediv__(self, other):
        try:
            return self._data / other._data
        except AttributeError:
            return self._data / other

    @_wrap_ret
    def __floordiv__(self, other):
        try:
            return self._data // other._data
        except AttributeError:
            return self._data // other

    @_wrap_ret
    def __mod__(self, other):
        try:
            return self._data % other._data
        except AttributeError:
            return self._data % other

    @_wrap_ret
    def __pow__(self, other, modulo=None):
        try:
            return self._data ** other._data
        except AttributeError:
            return self._data ** other

    @_wrap_ret
    def __matmul__(self, other):
        try:
            return self._data @ other._data
        except AttributeError:
            return self._data @ other

    @_wrap_ret
    def __and__(self, other):
        try:
            return self._data & other._data
        except AttributeError:
            return self._data & other

    @_wrap_ret
    def __or__(self, other):
        try:
            return self._data | other._data
        except AttributeError:
            return self._data | other

    @_wrap_ret
    def xor(self, other):
        try:
            return self._data ^ other._data
        except AttributeError:
            return self._data ^ other

    @_wrap_ret
    def __lshift__(self, other):
        return self._data << other

    @_wrap_ret
    def __rshift__(self, other):
        return self._data >> other

    # Comparison Operators
    @_wrap_ret
    def __eq__(self, other):
        try:
            return self._data == other._data
        except AttributeError:
            return self._data == other

    @_wrap_ret
    def __gt__(self, other):
        try:
            return self._data > other._data
        except AttributeError:
            return self._data > other

    @_wrap_ret
    def __lt__(self, other):
        try:
            return self._data < other._data
        except AttributeError:
            return self._data < other

    @_wrap_ret
    def __ge__(self, other):
        try:
            return self._data >= other._data
        except AttributeError:
            return self._data >= other

    @_wrap_ret
    def __le__(self, other):
        try:
            return self._data <= other._data
        except AttributeError:
            return self._data <= other


class DTensor(TorchArithmetic):
    """A dense multi-deimensional tensor on GPU (based on PyTorch).

    Parameters
    ----------
    value : np.ndarray or array_like
        Data of dense tensor.
    dtype : data-type, optional
        Data type of the tensor, by default None(inferred automatically)

    Examples
    --------
    You can create a `DTensor` in many ways. For example, `numpy.ndarray`
    `torch.Tensor` and `list of list`.

    >>> x = np.random.rand(3, 4)
    >>> A = st.DTensor.from_numpy(x)
    >>> A
    DTensor(
    tensor([[0.9604, 0.9878, 0.4680, 0.0225],
            [0.5677, 0.3412, 0.6834, 0.1586],
            [0.6883, 0.5613, 0.8819, 0.2513]], device='cuda:0',
        dtype=torch.float64)
    )
    >>> A = st.DTensor(x)
    >>> A
    DTensor(
    tensor([[0.9604, 0.9878, 0.4680, 0.0225],
            [0.5677, 0.3412, 0.6834, 0.1586],
            [0.6883, 0.5613, 0.8819, 0.2513]], device='cuda:0',
        dtype=torch.float64)
    )
    >>> x = torch.rand(3, 4)
    >>> A = st.DTensor(x)
    >>> A
    DTensor(
    tensor([[0.8155, 0.3988, 0.6346, 0.5665],
            [0.2955, 0.8131, 0.0774, 0.7335],
            [0.0919, 0.8163, 0.5959, 0.9401]], device='cuda:0')
    )
    >>> A = st.DTensor([[1, 2, 3, 4], [2, 1, 4, 3], [1, 3, 2, 4]])
    >>> A
    DTensor(
    tensor([[1, 2, 3, 4],
            [2, 1, 4, 3],
            [1, 3, 2, 4]], device='cuda:0')
    )
    """

    def __init__(self, data, dtype=None):
        if isinstance(data, DTensor):
            self._data = data._data
        elif isinstance(data, STensor):
            self._data = data._data.to_dense()
        else:
            self._data = torch.as_tensor(data, dtype=dtype)
        self._data: torch.Tensor = self._data.cuda()

    def __repr__(self):
        return '%s(\n%r\n)' % (type(self).__name__, self._data)

    def __len__(self):
        return self._data.__len__()

    def __copy__(self):
        return self._data.__copy__()

    def __iter__(self):
        return iter(self._data)

    # Slice
    @_wrap_ret
    def __getitem__(self, index):
        return self._data.__getitem__(index)

    def __setitem__(self, index, value):
        self._data.__setitem__(index, value)

    def __delitem__(self, index):
        self._data.__delitem__(index)

    @_wrap_ret
    def __getattr__(self, name):
        if name.startswith('_'):
            return super().__getattr__(name)
        ret = getattr(self._data, name)
        if callable(ret):
            raise AttributeError(
                f"{type(self).__name__} doesn't have attribute {name}")
        return ret

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            setattr(self._data, name, value)

    def __delattr__(self, name):
        if name.startswith('_'):
            super().__delattr__(name)
        else:
            delattr(self._data, name)

    @_wrap_ret
    def all(self, axis=None, keepdims=False):
        """Wrapper of `torch.all`

        Parameters
        ----------
        axis : None or int or tuple of ints , optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or bool
            Result of `all` operation as a DTensor or a boolean value.
        """
        if axis is None:
            return torch.all(self._data)
        else:
            return torch.all(self._data, dim=axis, keepdim=keepdims)

    @_wrap_ret
    def any(self, axis=None, keepdims=False):
        """Wrapper of `torch.any`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or bool
            Result of `any` operation as a DTensor or a boolean value.
        """
        if axis is None:
            return torch.any(self._data)
        return torch.any(self._data, dim=axis, keepdim=keepdims)

    @_wrap_ret
    def min(self, axis=None, keepdims=False):
        """Wrapper of `torch.Tensor.min`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `min` operation as a DTensor or a scalar value.
        """
        return self._data.min(dim=axis, keepdim=keepdims).values

    @_wrap_ret
    def max(self, axis=None, keepdims=False):
        """Wrapper of `torch.Tensor.max`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `max` operation as a DTensor or a scalar value.
        """
        return self._data.max(dim=axis, keepdim=keepdims).values

    @_wrap_ret
    def sum(self, axis=None, keepdims=False):
        """Wrapper of `torch.Tensor.sum`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `sum` operation as a DTensor or a scalar value.
        """
        if axis is None:
            return torch.sum(self._data)
        return self._data.sum(dim=axis, keepdim=keepdims)

    @_wrap_ret
    def prod(self, axis=None, keepdims=False):
        """Wrapper of `torch.Tensor.prod`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `prod` operation as a DTensor or a scalar value.
        """
        if axis is None:
            return torch.prod(self._data)
        return self._data.prod(dim=axis, keepdim=keepdims)

    @_wrap_ret
    def mean(self, axis=None, keepdims=False):
        """Wrapper of `torch.Tensor.mean`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `mean` operation as a DTensor or a scalar value.
        """
        if axis is None:
            return torch.mean(self._data)
        return self._data.sum(dim=axis, keepdim=keepdims)

    @_wrap_ret
    def var(self, axis=None, keepdims=False):
        """Wrapper of `torch.Tensor.var`, note that result is biased(divisor is `n`).

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are \
                left to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `var` operation as a DTensor or a scalar value.
        """
        if axis is None:
            return torch.var(self._data)
        return self._data.var(dim=axis, keepdim=keepdims)

    @_wrap_ret
    def std(self, axis=None, keepdims=False):
        """Wrapper of `torch.Tensor.std`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `std` operation as a DTensor or a scalar value.
        """
        if axis is None:
            return torch.std(self._data)
        return self._data.std(dim=axis, keepdim=keepdims)

    @_wrap_ret
    def dot(self, other):
        """Wrapper of `torch.Tensor.dot`. Perform tensor dot operation with
        another tensor, including vector inner product, matrix multiplication
        and general tensor dot.

        Parameters
        ----------
        other : DTensor
            The second operand of dot operation.

        Returns
        -------
        DTensor or scalar
            Tensor dot result as a DTensor or a scalar value.
        """
        if other.ndim == 1:
            return self._data.dot(other._data)
        else:
            return self._data.matmul(other._data)

    @_wrap_ret
    def reshape(self, shape):
        """Wrapper of `torch.Tensor.reshape`. Gives a new dense tensor with
        new shape and current data.

        Parameters
        ----------
        shape : int or tuple of ints
            Shape of new tensor.

        Returns
        -------
        DTensor
            New dense tensor with given shape.
        """
        return self._data.reshape(shape)

    @_wrap_ret
    def nonzero(self):
        """Wrapper of `torch.Tensor.nonzero`. Find the indices of non-zero elements.

        Returns
        -------
        tuple of DTensor
            Indices of non-zero elements.
        """
        return tuple([DTensor(d) for d in self._data.nonzero()])

    @_wrap_ret
    def astype(self, dtype):
        """Wrapper of `torch.Tensor.astype`. Convert the array to a new type.

        Parameters
        ----------
        dtype : data-type
            Type of new tensor.

        Returns
        -------
        DTensor
            New dense tensor with given data type.
        """
        return self._data.type(dtype)

    def to_numpy(self):
        """Return the tensor as `numpy.ndarray`.

        Returns
        -------
        numpy.ndarray
            Returned numpy array.
        """
        return self._data.cpu().numpy()

    @classmethod
    def from_numpy(cls, x):
        """Construct a `DTensor` from a `numpy.ndarray`.

        Parameters
        ----------
        x : numpy.ndarray
            Input array.

        Returns
        -------
        DTensor
            Dense tensor constructed from given array.

        Raises
        ------
        TypeError
            Raise if the parameter is not `numpy.ndarray`.
        """
        if not isinstance(x, np.ndarray):
            raise TypeError(
                f"Argument type should be `numpy.ndarray`, got {type(x)}")
        t = cls.__new__(cls)
        t._data = torch.from_numpy(x).cuda()
        return t


class STensor(TorchArithmetic):
    """A sparse multi-dimensional tensor on GPU (based on PyTorch).

    Parameters
    ----------
    data : Tuple of indices (ndim x nnz) and values (1 x nnz) or STenosr
        or DTensor
        Data of sparse tensor.
    shape : tuple, optional
        Shape of sparse tensor, by default None (inferred automatically)
    """

    def __init__(self, data, shape=None):
        if type(data) is tuple:
            indices, values = data
            self._data = tsparse.FloatTensor(indices, values, shape=shape)
        elif isinstance(data, DTensor):
            self._data = data._data.to_sparse()
        elif isinstance(data, STensor):
            self._data = data._data
        else:
            self._data = tsparse.FloatTensor(data, shape=shape)
        self._data = self._data.cuda().coalesce()

    def __repr__(self):
        return '%s(\n%r\n)' % (type(self).__name__, self._data)

    def __len__(self):
        return self._data.__len__()

    def __copy__(self):
        return self._data.__copy__()

    def __iter__(self):
        return iter(self._data)

    # Notice that Pytorch's GPU sparse tensor doesn't support slice fully yet!
    # Slice
    @_wrap_ret
    def __getitem__(self, index):
        return self._data.__getitem__(index)

    def __setitem__(self, index, value):
        self._data.__setitem__(index, value)

    def __delitem__(self, index):
        self._data.__delitem__(index)

    @_wrap_ret
    def __getattr__(self, name):
        if name.startswith('_'):
            return super().__getattr__(name)
        ret = getattr(self._data, name)
        if callable(ret):
            raise AttributeError(
                f"{type(self).__name__} doesn't have attribute {name}")
        return ret

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            setattr(self._data, name, value)

    def __delattr__(self, name):
        if name.startswith('_'):
            super().__delattr__(name)
        else:
            delattr(self._data, name)

    @_wrap_ret
    def all(self, axis=None, keepdims=False):
        """Wrapper of `torch.all`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or bool
            Result of `all` operation as a STensor or a boolean value.
        """
        if keepdims:
            warnings.warn(f"The keepdims parameter has no effect yet",
                          UserWarning)
        if axis is None:
            return torch.all(self._data)
        return torch.all(self._data, axis=axis, keepdim=keepdims)

    @_wrap_ret
    def any(self, axis=None, keepdims=False):
        """Wrapper of `torch.any`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or bool
            Result of `any` operation as a STensor or a boolean value.
        """
        if keepdims:
            warnings.warn(f"The keepdims parameter has no effect yet",
                          UserWarning)
        if axis is None:
            return torch.any(self._data)
        return torch.any(self._data, axis=axis, keepdim=keepdims)

    @_wrap_ret
    def min(self, axis=None, keepdims=False):
        """Not implmented yet.

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `min` operation as a STensor or a scalar value.
        """
        raise NotImplementedError

    @_wrap_ret
    def max(self, axis=None, keepdims=False):
        """Not implemented yet.

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `max` operation as a STensor or a scalar value.
        """
        raise NotImplementedError

    @_wrap_ret
    def sum(self, axis=None, keepdims=False):
        """Wrapper of `torch.sparse.sum`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `sum` operation as a STensor or a scalar value.
        """
        if keepdims:
            warnings.warn(f"The keepdims parameter has no effect yet",
                          UserWarning)
        if axis is None:
            return tsparse.sum(self._data)
        return tsparse.sum(self._data, dim=axis)

    @_wrap_ret
    def prod(self, axis=None, keepdims=False):
        """Not implemented yet.

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `prod` operation as a STensor or a scalar value.
        """
        raise NotImplementedError

    @_wrap_ret
    def mean(self, axis=None, keepdims=False):
        """Not implemented yet.

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `mean` operation as a STensor or a scalar value.
        """
        raise NotImplementedError

    @_wrap_ret
    def var(self, axis=None, keepdims=False):
        """Not implemented yet.

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `var` operation as a STensor or a scalar value.
        """
        raise NotImplementedError

    @_wrap_ret
    def std(self, axis=None, keepdims=False):
        """Not implemented yet.

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default
            None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `std` operation as a STensor or a scalar value.
        """
        raise NotImplementedError

    @_wrap_ret
    def dot(self, other):
        """Wrapper of `torch.sparse.mm`. Perform tensor dot operation with
        another tensor, including vector inner product, matrix multiplication
        and general tensor dot.

        Parameters
        ----------
        other : DTensor or STensor
            The second operand of dot operation.

        Returns
        -------
        DTensor or STensor
            Tensor dot result as a DTensor or a STensor or a scalar value.
        """
        return tsparse.mm(self._data, other._data)

    @_wrap_ret
    def todense(self):
        """Convert to a dense tensor.

        Returns
        -------
        DTensor
            Dense tensor coverted from the sparse tensor.
        """
        return self._data.to_dense()

    @_wrap_ret
    def reshape(self, shape):
        """Not implemented yet.

        Parameters
        ----------
        shape : int or tuple of ints
            Shape of new tensor.

        Returns
        -------
        STensor
            New sparse tensor with given shape.
        """
        raise NotImplementedError

    @_wrap_ret
    def nonzero(self):
        """Wrapper of `torch.Tensor.indices`. Find the indices of non-zero elements.

        Returns
        -------
        tuple of DTensor
            Indices of non-zero elements.
        """
        return tuple([x for x in self._data.indices()])

    @_wrap_ret
    def astype(self, dtype):
        """Not implemented yet.

        Parameters
        ----------
        dtype : data-type
            Type of new tensor.

        Returns
        -------
        STensor
            New sparse tensor with given data type.
        """
        raise NotImplementedError

    def to_scipy(self, format='coo'):
        """Return the sparse tensor as a `scipy.sparse.spmatrix`.

        Parameters
        ----------
        format : str, optional, {'coo', 'csr', 'csc', 'lil', 'dok'}
            Format of scipy sparse matrix, by default 'coo'

        Returns
        -------
        scipy.sparse.spmatrix
            Returned scipy sparse matrix.
        """
        if self._data.ndim != 2:
            raise ValueError(
                f"Only 2-dim sparse tensor can be converted to scipy sparse\
                matrix, got {self._data.ndim}")
        indices = self._data.indices().cpu().numpy()
        row, col = indices[0], indices[1]
        vals = self._data.values().cpu().numpy()
        shape = tuple(self._data.shape)
        return ssp.coo_matrix((vals, (row, col)), shape=shape).asformat(format)

    def to_sparse_array(self):
        """Return the sparse tensor as a `sparse.COO`.

        Returns
        -------
        sparse.COO
            Returned sparse array.
        """
        indices = self._data.indices().cpu().numpy()
        vals = self._data.values().cpu().numpy()
        shape = tuple(self._data.shape)
        return sparse.COO(indices, vals, shape=shape)

    @classmethod
    def from_numpy(cls, x: np.ndarray):
        """Construct a `STensor` from a `numpy.ndarray`.

        Parameters
        ----------
        x : numpy.ndarray
            Input array.

        Returns
        -------
        STensor
            Sparse tensor constructed from given array.

        Raises
        ------
        TypeError
            Raise if the parameter is not `numpy.ndarray`.
        """
        if not isinstance(x, np.ndarray):
            raise TypeError(
                f"Argument type should be `numpy.ndarray`, got {type(x)}")
        t = cls.__new__(cls)
        t._data = torch.from_numpy(x).to_sparse().cuda()
        t._data = t._data.coalesce()
        return t

    @classmethod
    def from_scipy_sparse(cls, x: scipy.sparse.spmatrix):
        """Construct a `STensor` from a `scipy.sparse.spmatrix`.

        Parameters
        ----------
        x : scipy.sparse.spmatrix
            Input sparse matrix.

        Returns
        -------
        STensor
            Sparse tensor constructed from given sparse matrix.

        Raises
        ------
        TypeError
            Raise if the parameter is not `scipy.sparse.spmatrix`.
        """
        if not isinstance(x, scipy.sparse.spmatrix):
            raise TypeError(
                f"Argument type should be `scipy.sparse.spmatrix`, \
                got {type(x)}")
        x = x.tocoo()
        indices = torch.from_numpy(np.vstack([x.row, x.col])).type(torch.long)
        values = torch.from_numpy(x.data)
        shape = torch.Size(x.shape, dtype=torch.long)
        t = cls.__new__(cls)
        t._data = tsparse.FloatTensor(indices, values, shape).cuda()
        t._data = t._data.coalesce()
        return t

    @classmethod
    def from_sparse_array(cls, x: sparse.COO):
        """Construct a sparse tensor from a `sparse.COO`.

        Parameters
        ----------
        x : sparse.COO
            Input COO sparse array.

        Returns
        -------
        STensor
            Sparse tensor constructed from given sparse array.

        Raises
        ------
        TypeError
            Raise if the parameter is not `sparse.COO`.
        """
        if not isinstance(x, sparse.COO):
            raise TypeError(
                f"Argument type should be `sparse.SparseArray`, got {type(x)}")
        indices = torch.from_numpy(x.coords).type(torch.long)
        values = torch.from_numpy(x.data)
        shape = torch.Size(x.shape, dtype=torch.long)
        t = cls.__new__(cls)
        t._data = tsparse.FloatTensor(indices, values, shape).cuda()
        t._data = t._data.coalesce()
        return t

    @classmethod
    def from_torch(cls, x: torch.Tensor):
        """Construct a `STensor` from a `torch.Tensor`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        STensor
            Sparse tensor constructed from given array.

        Raises
        ------
        TypeError
            Raise if the parameter is not `torch.Tensor`.
        """
        if not isinstance(x, torch.Tensor):
            raise TypeError(
                f"Argument type should be `torch.Tensor`, got {type(x)}")
        t = cls.__new__(cls)
        if not x.is_sparse:
            x = x.to_sparse()
        t._data = x.cuda()
        t._data = t._data.coalesce()
        return t



from .._model import DMmodel
from .res_util import *
from ...util.basicutil import param_default
from .. import CubeFlow


class CubeFlowPlus(DMmodel):
    def __init__(self, tensorList:list):
        self.amt_tensor, self.cmt_tensor = tensorList[0], tensorList[1]
    
    def __str__(self):
        return str(vars(self))
    
    def run(self, **params):
        self.res = param_default(params, 'cf_res', None)
        self.handle_biggraph_type = param_default(params, 'handle_biggraph_type', 1)
        self.max_node_limit = param_default(params, 'max_node_limit', 100)
        self.maxsize1 = param_default(params, 'maxsize1', -1)
        self.maxsize2 = param_default(params, 'maxsize2', (-1, 100, -1))
        self.outpath1 = param_default(params, 'outpath1', '') # CubeFlow's result without limit
        self.outpath2 = param_default(params, 'outpath2', '') 
        self.alpha = param_default(params, 'alpha', 0.8)
        self.k = param_default(params, 'k', 1)
        self.dim = param_default(params, 'dim', 3)
        self.del_type = param_default(params, 'del_type', 1)
        self.is_find_all_blocks = param_default(params, 'is_find_all_blocks', False)
        
        self.initData()
        
        if self.res is None:    
            self.res = self.call_cubeflow()

        # 步骤一：划分大图为连通子图
        subgraph_connected_list, subgraph_timebin_list, subgraph_timebin_idx_list, subgraph_can_divided = \
            divide_connected_conponents(self.res, self.amt_tensor, self.cmt_tensor)

        # 步骤二：对每个大图（总节点数超过最大限制）进行单独处理
        # 选择1：去掉度最大的节点，放入连通图算法
        # 选择2：运行带约束的CubeFlow，直至跑空所有节点为止

        subgraph_connected_list, subgraph_timebin_list, subgraph_timebin_idx_list = \
            handle_big_graph(subgraph_connected_list, subgraph_timebin_list, subgraph_timebin_idx_list,
                             subgraph_can_divided, self.max_node_limit,
                             self.amt_tensor, self.cmt_tensor, self.handle_biggraph_type, self.maxsize2, self.alpha, self.dim, self.del_type)

        print(len(subgraph_connected_list))

        # 步骤三：对连通子图列表打分并排序

        subgraph_score_list, subgraph_score1_list, subgraph_score2_list, amct_list = \
            score_connected_graph_list(subgraph_connected_list, subgraph_timebin_list, subgraph_timebin_idx_list,
                                       self.alpha, self.amt_stensor, self.cmt_stensor, self.amt_tensor, self.cmt_tensor)

        subgraph_score_list = np.array(subgraph_score_list)
        subgraph_score_sort_index_list = np.argsort(subgraph_score_list)[::-1]  # 分数由高到低排序
        res_new = []

        for idx in subgraph_score_sort_index_list:
            res_new.append([amct_list[idx], subgraph_score_list[idx]])
            
        if self.outpath2 != '':
            for i in range(len(res_new)):
                saveres2txt(res_new[i], self.outpath2)
    
            print(f'save new result to path:{self.outpath2} success!')
        
        return res_new, subgraph_score_list, subgraph_score1_list, subgraph_score2_list, amct_list           
        
    def initData(self):
        self.amt_stensor = self.amt_tensor.toSTensor(hasvalue=True)
        self.cmt_stensor = self.cmt_tensor.toSTensor(hasvalue=True)
        print(self.amt_stensor.shape)
        print(self.cmt_stensor.shape)
        
    
    def call_cubeflow(self):
        cf = CubeFlow([self.amt_stensor, self.cmt_stensor], alpha=self.alpha, k=self.k, dim=self.dim, outpath=self.outpath1)
        res = cf.run(del_type=self.del_type, maxsize=self.maxsize1, is_find_all_blocks=self.is_find_all_blocks)
        return res
    
    
        
        
    



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  (Discrete) Normal Gaussian distribution
#  Author: wenchieh
#
#  Project: eaglemine
#      normal_gaussian.py
#      Version:  1.0
#      Date: December 05 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/5/2017>
#

__author__ = 'wenchieh'

# third-party lib
import numpy as np
from scipy.stats import multivariate_normal

# project
from .norm_extras import mvnormcdf


class NormalGaussian(object):
    def __init__(self):
        self._mu, self._cov = None, None
        self._ndim = 0

    def set_para(self, mu, cov):
        self._mu = np.array(mu)
        self._ndim = len(mu)
        self._cov = np.array(cov)

    def pdf(self, xs):
        return multivariate_normal.pdf(xs, self._mu, self._cov, allow_singular=True)

    def cdf(self, xs):
        return mvnormcdf(xs, self._mu, self._cov)

    def range_cdf(self, left, right):
        return mvnormcdf(right, self._mu, self._cov, left)


#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#    feature2histogram.py
#      Version:  1.0
#      Goal: Routine scripts
#      Created by @wenchieh  on <12/27/2018>


__author__ = 'wenchieh'

# sys
import argparse

# third-party
import numpy as np
from scipy.sparse import csr_matrix

# project
from .utils.loader import Loader
from .tools.histogram_heuristic_generator import HistogramHeuristicGenerator
from .utils.ploter import plot_heatmap, plot_heatmap_graphlab_pgrk, plot_heatmap_2discretes


VERBOSE = True

def histogram_construct(feats, degidx, outs_hist, outs_node2hcel, outs_hcel2avgfeat, mode=2):
    n_samples, n_features = feats.shape
    index = np.array([True] * n_samples)
    for mod in range(mode):
        index &= feats[:, mod] > 0
    if VERBOSE:
        print("total shape: {}, valid samples:{}".format(feats.shape, np.sum(index)))

    degree, features = None, None
    feats = feats[index, :]
    if degidx > 0:
        degree = feats[:, degidx - 1]
        features = np.delete(feats, degidx - 1, axis=1)
        del feats
    else:
        features = feats

    hist_gen = HistogramHeuristicGenerator()
    if degree is not None:
        hist_gen.set_deg_data(degree, features)
        hist_gen.histogram_gen(method="degree", N=80, base=10)
    else:
        n_buckets = 80
        logarithmic, base = True, 10
        hist_gen.set_data(features)
        hist_gen.histogram_gen(method="N", N=n_buckets, logarithmic=logarithmic, base=base)

    if VERBOSE:
        hist_gen.dump()

    hist_gen.save_histogram(outs_hist)
    hist_gen.save_pts_index(outs_node2hcel, pts_idx=np.arange(n_samples)[index])
    hist_gen.save_hpos2avgfeat(outs_hcel2avgfeat)
    print("done!")

def histogram_view(ins_hist, x_lab, y_lab, outs_viz=None):
    loader = Loader()
    _shape_, ticks_vec, hist_arr = loader.load_multi_histogram(ins_hist)
    csr_mat = csr_matrix((hist_arr[:, -1], (hist_arr[:, 0], hist_arr[:, 1])), shape=_shape_, dtype=int)
    plot_heatmap(ticks_vec[1], ticks_vec[0], csr_mat.toarray(), xlabel=x_lab, ylabel=y_lab, outfn=outs_viz)
    # plot_heatmap_graphlab_pgrk(ticks_vec[1], ticks_vec[0], csr_mat.toarray(), xlabel=x_lab, ylabel=y_lab, outfn=outs_viz)
    # plot_heatmap_2discretes(ticks_vec[1], ticks_vec[0], csr_mat.toarray(), xlabel=x_lab, ylabel=y_lab, outfn=outs_viz)
    print('done!')


def demo():
    mode = 2
    outs = '../output/'
    ins_gfeat = '../example/outd2hub_feature'
    outs = '../output/'
    ofn_hist = 'histogram.out'
    ofn_node2hcel = 'node2hcel.out'
    ofn_hcel2avgfeat = 'hcel2avgfeat.out'
    ofn_heatmap = 'heatmap.png'
    x_lab, y_lab = ["Hubness", "Out-degree"] # ["Authority", "In-degree"], ["PageRank", "Degree"], ["Degree", "Triangles"]

    loader = Loader()
    m, _, gfts = loader.load_features(ins_gfeat, float)
    histogram_construct(gfts[:m], 1, outs + ofn_hist, outs + ofn_node2hcel, outs + ofn_hcel2avgfeat, mode)
    histogram_view(outs + ofn_hist, x_lab, y_lab, outs + ofn_heatmap)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Construct histogram for given TWO-dimension feature",
                                     usage="python feature2histogram.py [ins], [degidx] [x_idx] [y_idx] [x_lab] [y_lab]"
                                           "[outs_hist] [outs_node2hcel] [outs_hcel2avgfeat] [outs_viz] [delimiter] [comments]")
    parser.add_argument("ins", type=str, help="input feature path")
    parser.add_argument("degidx", type=int, default=0, help="feature index if contain (in/out-) degree for graph else 0")
    parser.add_argument("x_idx", type=int, help="feature index for x axis")
    parser.add_argument("y_idx", type=int, help="feature index for y axis")
    parser.add_argument("x_lab", type=str, help="feature label for x axis")
    parser.add_argument("y_lab", type=str, help="feature label for y axis")
    parser.add_argument("outs_hist", type=str, help="output path for result of feature-to-histogram ")
    parser.add_argument("outs_node2hcel", type=str, help="output path for result of point to cell of histogram")
    parser.add_argument("outs_hcel2avgfeat", type=str, help="output path for result of average feature of each cell in histogram")
    parser.add_argument("outs_viz", type=str, help="output path for result of histogram view")
    parser.add_argument("delimiter", type=str, default=',', help="delimiter of the input")
    parser.add_argument("comments", type=str, default='%', help="comments character of the input")
    args = parser.parse_args()

    mode = 2
    feat = np.loadtxt(args.ins, float, args.comments, args.delimiter, usecols=[args.x_idx, args.y_idx])
    histogram_construct(feat, args.degidx, args.outs_hist, args.outs_node2hcel, args.outs_hcel2avgfeat, mode)
    histogram_view(args.outs_hist, args.x_lab, args.y_lab, args.outs_viz)



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#    ploter_aux.py
#		  auxiliary tools for visualization 
#      Version:  1.0
#      Goal: Subroutine script
#      Created by @wenchieh  on <11/28/2017>


__author__ = 'wenchieh'

import warnings
warnings.filterwarnings("ignore")

# third-party lib
import numpy as np
from numpy.linalg import eig
import matplotlib
matplotlib.use("Agg")
import matplotlib.pylab as plt
from matplotlib.patches import Ellipse
from matplotlib.colors import LogNorm


def _covmat_angle_(cov):
    cov = np.asarray(cov).reshape((2, 2))
    _, v = eig(cov)
    cosine_theta = v[:, 0].dot(np.array([0, 1])) * 1.0 / np.sqrt(np.sum(v[:, 0]**2))
    theta = np.arccos(cosine_theta) * 180 / np.pi
    return theta


def _construct_ellipses_(clusters_para, scale=5.0, SINGLE_PARAS=5):
    N_m = len(clusters_para)

    ell, mus = list(), list()
    left, right, bottom, up = 0, 0, 0, 0
    for k in range(N_m):
        norm = 1.0
        para = clusters_para[k]
        mu = para[:2][::-1]
        cov = np.array([[para[2], para[3]],[para[3], para[4]]])

        angle = _covmat_angle_(cov)
        w, h = scale * np.sqrt(cov[0, 0]) / norm, scale * np.sqrt(cov[1, 1]) / norm
        if w + mu[0] > right:
            right = w + mu[0]
        if h + mu[1] > up:
            up = h + mu[1]
        if -h + mu[1] < bottom:
            bottom = -h + mu[1]

        if mu[1] < 0:    # for  better illustration
            w *= 1.5
            h *= 1.5

        mus.append(mu)
        ell.append(Ellipse(mu, w, h, angle))

    return ell, mus, [left, right, bottom, up]

## ellipses labeled heatmap plot
def plot_heatmap_ellipse(x_vec, y_vec, heatmap, clusters_para, scale=5, base=10,
                         cmap='gray_r', xlabel=None, ylabel=None, outfn=None):
    n, m = heatmap.shape
    fig = plt.figure()
    plt.pcolormesh(heatmap, cmap=cmap, norm=LogNorm(), rasterized=True)
    cb = plt.colorbar()
    for l in cb.ax.yaxis.get_ticklabels():
        l.set_family('Times New roman')
        l.set_size(20)

    ell, mus, bnds = _construct_ellipses_(clusters_para, scale)
    mus = np.asarray(mus)

    ax = fig.gca()
    for k in range(len(ell)):
        e = ell[k]
        ax.add_artist(e)
        e.set_clip_box(ax.bbox)
        e.set_facecolor(None)
        e.set_fill(False)
        e.set_linewidth(2)
        e.set_color('#ED1C24')
        e.set_linestyle('solid')

    plt.plot(mus[:, 0], mus[:, 1], 'k+', markersize=10, linewidth=5)

    nw_xtick = ['%dE%s' % (base, int(x_vec[int(xt)])) if xt < m else '' for xt in ax.get_xticks()]
    nw_ytick = [r'$%d$' % int(np.power(base, y_vec[int(yt)])) if yt < n else '' for yt in ax.get_yticks()]
    if nw_xtick[-1] == '':
        nw_xtick[-1] = '%.2f'%np.power(base, x_vec[-1])
    if nw_ytick[-1] == '':
        nw_ytick[-1] = '%d' % int(y_vec[-1])
        nw_ytick = '%d' % int(np.power(base, y_vec[-1]))

    ax.set_xticklabels(nw_xtick, fontsize=20)
    ax.set_yticklabels(nw_ytick, fontsize=20)

    if xlabel is not None:
        plt.xlabel(xlabel, linespacing=12, fontsize=32, family='Times New roman')
    if ylabel is not None:
        plt.ylabel(ylabel, linespacing=12, fontsize=32, family='Times New roman')

    fig.set_size_inches(8, 7)
    fig.tight_layout()
    if outfn is not None:
        plt.savefig(outfn)
        plt.close()

    return fig

## ellipses (two scale 3 & 5) labeled heatmap plot
def plot_heatmap_ellipse_covs(x_vec, y_vec, heatmap, paras, base=10, scales=(3, 5),
                         cmap='gray_r', xlabel=None, ylabel=None, outfn=None):
    n, m = heatmap.shape
    fig = plt.figure(figsize=(8, 6.5), dpi=96)
    plt.pcolormesh(heatmap, cmap=cmap, norm=LogNorm(), rasterized=True)
    cb = plt.colorbar()
    for l in cb.ax.yaxis.get_ticklabels():
        l.set_family('Times New roman')
        l.set_size(20)

    ax = fig.gca()
    ell_s1, mus, bnds = _construct_ellipses_(paras, scales[0])
    ell_s2, _, _ = _construct_ellipses_(paras, scales[1])
    ell = list(ell_s1)
    ell.extend(ell_s2)

    for k in range(len(ell)):
        e = ell[k]
        ax.add_artist(e)
        e.set_clip_box(ax.bbox)
        # e.set_alpha(rnd.rand())
        e.set_facecolor(None)
        e.set_fill(False)
        e.set_linewidth(1.5)
        # e.set_edgecolor(rnd.rand(3))
        # e.set_edgecolor(COLORS[k])
        if k < len(ell_s1):
            # e.set_linewidth(3)
            # e.set_color('#ED1C24')
            e.set_color('#FFF200')
            e.set_linestyle('dashed')
        else:
            # e.set_linewidth(2)
            # e.set_color('#2704FB')
            e.set_color('#F7030C')
            e.set_linestyle('solid')

    xticks = ax.get_xticks()
    if xticks[-1] > m:  xticks = xticks[:-1]

    nw_xtick = ['%dE%d' % (base, int(np.log(x_vec[int(xt)]) / np.log(base)))
                if ((xt < m) and (xt % 20 == 0)) else '' for xt in xticks]
    # nw_ytick = ['%d' % int(y_vec[int(yt)]) if yt < n else '' for yt in ax.get_yticks()]
    nw_ytick = []
    for yt in ax.get_yticks():
        if yt < n:
            yval = y_vec[int(yt)]
            if yval < 1e4:
                nw_ytick.append(r'%d' % yval)
            else:
                pws = int(np.log10(yval))
                fv = yval * 1.0 / 10**pws
                nw_ytick.append('%.1fE%d'%(fv, pws))

    if nw_xtick[-1] == '':
        nw_xtick[-1] = '%.2f' % x_vec[-1]
        # nw_xtick[-1] = '%.2f'%np.power(base, x_vec[-1])
    if nw_ytick[-1] == '':
        nw_ytick[-1] = r'$%d$' % int(y_vec[-1])
        # nw_ytick = '%d' % int(np.power(base, y_vec[-1]))

    ax.set_xticklabels(nw_xtick, fontsize=23, family='Times New roman')   # , fontweight='bold'
    ax.set_yticklabels(nw_ytick, fontsize=23, family='Times New roman')

    if xlabel is not None:
        plt.xlabel(xlabel, fontsize=32, family='Times New roman') #, fontweight='bold'
    if ylabel is not None:
        plt.ylabel(ylabel, fontsize=32, family='Times New roman')

    # fig.set_size_inches(8, 7)
    fig.tight_layout()
    if outfn is not None:
        plt.savefig(outfn)
        plt.close()

    return fig



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  Describe the milti-dimensional islands (hypercube centered) with gaussian model.
#  Author: wenchieh
#
#  Project: eaglemine
#      gaussian_describe.py
#      Version:  1.0
#      Date: December 12 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/12/2017>
#

__author__ = 'wenchieh'

# third-party lib
import numpy as np

# project
from .discrete_gaussian import DiscreteGaussian


class GaussianDescribe(object):
    _SERIALIZE_DELIMITER_ = ';'
    def __init__(self, ndim=2, is_mix=False, n_components=2):
        self.descriptor = DiscreteGaussian()
        self.is_mix = is_mix
        self.ndim = ndim
        self.paras = None
        self._npos = None
        self.data = None
        self.values = None
        self._total_values = None
        self.n_components = n_components if is_mix else 1

    def set_data(self, hypercubes, values):
        self.data = np.asarray(hypercubes, float)
        self.values = np.asarray(values)
        # self.values = np.asarray(values, int)
        self._npos, self._total_values = len(self.data), np.sum(self.values)

    def fit(self, hypercubes, values, type='left'):
        """
        fit the hypercubes with edge-length = 1 with Gaussian descriptor
        :param hypercubes: input hypercubes,  which dimension >= 2
        :param values: value in each hypercube (maybe #point or weight)
        :param type: indicate the hypercube position,  "center" | "left" | "right"
                     The coordinate given in hypercube is at ["center" | "left" | "right"] of affiliated hypercubes
        :return:
        """
        self.set_data(hypercubes, values)
        self.ndim = np.ndim(self.data)

        left, right = self.data, self.data + 1
        if type == 'center':
            left, right = self.data - 0.5, self.data + 0.5
        elif type == 'right':
            left, right = self.data - 1, self.data
        else:
            pass

        if self.is_mix:
            mus, covs, ws, loss = self.descriptor.fit_mixture(left, right, self.values, self.n_components)
        else:
            mus, covs, ws, loss = self.descriptor.fit_single(left, right, self.values)

        self.paras = {'mus': mus, 'covs':covs, 'weights':ws, 'loss':loss}
        # print("data fit done!")

    def apply(self, hypercubes, values, type='left', loss_update=False):
        values = np.asarray(values, int)
        if self.paras is not None:
            pos = np.asarray(hypercubes)
            m, n = pos.shape
            if n != self.ndim:
                RuntimeWarning("Input data dimension iss not consistence with initialization!")
                if n > self.ndim:
                    raise ValueError("Input data dimension ({}) exceed initial dimension ({})".format(n, self.ndim))
            else:
                left, right = pos, pos + 1
                if type == 'center':
                    left, right = pos - 0.5, pos + 0.5
                elif type == 'right':
                    left, right = pos - 1, self.data
                else:
                    pass
                loss = self.descriptor.log_loss(self.paras, left, right, values, self.is_mix, self.n_components)
                if loss_update:
                    self.paras["loss"] = loss
                return loss
        else:
            raise RuntimeError("No derived fitting parameters!")

    def dump(self):
        if self.paras is not None:
            print("descriptor information:")
            print("#dimension: {}, #pos: {}, total count: {}".format(self.ndim, self._npos, self._total_values))
            print("parameters: {}".format("mixture(#{})".format(self.n_components) if self.is_mix else "single"))
            print("\t mus:{}, covs: {}, weights: {}, loss: {}".format(self.paras["mus"], self.paras["covs"],
                                                                      self.paras["weights"], self.paras["loss"]))
        else:
            raise RuntimeError("No data fit")
        print("done!\n")

    def __str__(self):
        content = ""
        if self.paras is not None:
            mixture = 1 if self.is_mix is True else 0
            mus_lst, covs_lst = list(), list()
            for mu in self.paras["mus"]:
                mus_lst.append(np.asarray(mu).tolist())
            for cov in self.paras["covs"]:
                covs_lst.append(np.asarray(cov).tolist())
            ws_str = str(np.asarray(self.paras["weights"]).tolist())
            content += str(self.ndim) + self._SERIALIZE_DELIMITER_ + str(mixture) + \
                       self._SERIALIZE_DELIMITER_ + str(mus_lst) + self._SERIALIZE_DELIMITER_ + \
                       str(covs_lst) + self._SERIALIZE_DELIMITER_ + ws_str + self._SERIALIZE_DELIMITER_ + \
                       str(self.paras["loss"]) + self._SERIALIZE_DELIMITER_ + str(len(self.data)) + \
                       self._SERIALIZE_DELIMITER_ + str(np.sum(self.values))
        return content

    def load(self, desc_str, verbose=False):
        toks = desc_str.strip().split(self._SERIALIZE_DELIMITER_)
        self.ndim = int(toks[0])
        mixture = int(toks[1])
        self.is_mix = True if mixture == 1 else False
        self.paras = dict()
        self.paras["mus"] = np.asarray(eval(toks[2]))
        self.paras["covs"] = np.asarray(eval(toks[3]))
        self.paras["weights"] = np.asarray(eval(toks[4]))
        self.paras["loss"] = float(toks[5])
        self._npos = int(toks[6])
        self._total_values = float(toks[7])
        self.n_components = len(self.paras["mus"])
        if verbose:
            self.dump()

    def compact_parm(self):
        parms = list()
        parms.extend(list(np.asarray(self.paras['mus']).squeeze().flatten()))
        if self.is_mix:
            parms.extend(list(self.paras['weights'][:-1]))
        for k in range(self.n_components):
            cov = self.paras['covs'][k]
            for m in range(self.ndim):
                parms.extend(cov[m, m:])
        return parms


#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  Histogram generator with heuristic rules for eliminating black degree lines for degree sequence.
#  Author: wenchieh
#
#  Project: eaglemine
#      histogram_heuristic_generator.py
#      Version:  1.0
#      Date: December 17 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/17/2017>
# 

__author__ = 'wenchieh'

# sys
import bisect

# third-party lib
import numpy as np


class HistogramHeuristicGenerator(object):
    def __init__(self):
        self._mode = 0
        self._npts = 0
        self._features = None
        self._ticks_vec = None
        self.shape = None
        self.histogram = None
        self.points_coord = None
        self.hpos2avgfeat = None

    def set_deg_data(self, degrees, features):
        assert(len(degrees) == len(features))
        # self._features = np.hstack([np.array(degrees).T[:, np.newaxis], np.asarray(features).T[:, np.newaxis]])
        self._features = np.column_stack((degrees, features))
        self._npts, self._mode = self._features.shape

    def set_data(self, features):
        self._features = np.array(features)
        self._npts, self._mode = self._features.shape

    def _bucketize_N_(self, values, N, logarithmic=False, base=10):
        values = np.array(values)
        if logarithmic:
            if values.min() < 0:
                values = values[values >= 0]
            if values.min() == 0:
                values += 1.0
            values = np.log(values) / np.log(base)

        min_val, max_val = values.min(), values.max()
        bandwidth = (max_val - min_val) * 1.0 / N
        bins_vec = np.array([min_val + i * bandwidth for i in range(N + 1)])

        if logarithmic:
            bins_vec = np.power(base, bins_vec)

        return bins_vec

    def _bucketize_logP_(self, values, P):
        logP = np.log(P)
        values = np.array(values)
        values = values[values > 0]
        min_lgpv, max_lgpv = np.log(values.min()) * 1.0 / logP, np.log(values.max()) * 1.0 / logP
        N_bins = max_lgpv - min_lgpv + 1
        bins_vec = np.array(np.power(P, min_lgpv + i * logP) for i in range(N_bins))

        return bins_vec

    def _bucketize_degree_(self, degrees, base=10):
        start, step = 0, 0
        if base == 10:
            start, step = 10, 0.05
        elif base == 2:
            start, step = 16, 0.1
        else:
            print("Error exception: Not a proper step setting")
            exit()

        sparse_seq = range(1, start + 1)
        min_lgdeg = np.log(np.max(sparse_seq)) * 1.0 / np.log(base)
        max_lgdeg = np.log(np.max(degrees + 1)) * 1.0 / np.log(base)
        init_dense_seq = np.power(base, np.arange(min_lgdeg + step, max_lgdeg + step, step))
        init_dense_seq = np.asarray(init_dense_seq, int)
        dense_seq = list()
        dense_seq.extend(init_dense_seq[:2])

        for i in range(2, len(init_dense_seq)):
            pre_bw = int(dense_seq[-1] - dense_seq[-2])
            cur = init_dense_seq[i]
            if int(cur - dense_seq[-1]) < pre_bw:
                dense_seq.append(dense_seq[-1] + pre_bw)
            else:
                dense_seq.append(cur)

        bins_vec = np.hstack([sparse_seq, np.asarray(dense_seq)])

        return bins_vec

    def _seq2hists_(self, seq, ticks):
        Npts, nbins = len(seq), len(ticks) - 1
        N_outliers = 0

        hist = np.array([0] * nbins, int)
        pts_pos = np.array([0] * Npts, int)

        for pt_idx in range(Npts):
            pt_val = seq[pt_idx]
            if pt_val < ticks[0]:
                N_outliers += 1
                pts_pos[pt_idx] = -1
            else:
                k_idx = bisect.bisect_left(ticks, pt_val)
                k_idx = min([k_idx, nbins])
                if k_idx > 0:
                    if abs(ticks[k_idx] - pt_val) > 1e-14:
                        k_idx = max([0, k_idx - 1])
                    k_idx = min([k_idx, nbins - 1])
                hist[k_idx] += 1
                pts_pos[pt_idx] = k_idx

        return hist, pts_pos, N_outliers

    def _check_data_(self):
        if np.min(self._features) <= 0.0:
            ValueError("Error: The input feature contains non-positive values, which can't be applied with logarithmic")

    def _degree_feature_bucketize_(self, base):
        if self._ticks_vec is None:
            self._ticks_vec = list()
        if self.shape is None:
            self.shape = list()

        deg_bucks = self._bucketize_degree_(self._features[:, 0], base)
        self._ticks_vec.append(deg_bucks)
        self.shape.append(len(deg_bucks))
        for mod in range(1, self._mode):
            values = self._features[:, mod]
            values = values[values > 0]
            min_lgpv, max_lgpv = np.log(values.min()) * 1.0 / np.log(base), np.log(values.max()) * 1.0 / np.log(base)
            N_bins = int(np.ceil(len(deg_bucks) / 10.0)) * 10
            bucks = min_lgpv + (max_lgpv - min_lgpv) * np.arange(N_bins + 1) * 1.0 / N_bins
            self._ticks_vec.append(np.power(base, bucks))
            self.shape.append(len(bucks))

    def _multidiscrete_bucketize_(self, discretes_index, base):
        '''
        bucketize for multiple discrete features in log scale with base=base, like degree, # triangle, etc.
        :param discretes_index: discrete features index in data
        :param base: the base of log
        '''
        if self._ticks_vec is None:
            self._ticks_vec = list()
        if self.shape is None:
            self.shape = list()

        max_disc_bins = 0
        for mod in range(self._mode):
            values = self._features[:, mod]
            values = values[values > 0]
            if mod in discretes_index:
                disc_bucks = self._bucketize_degree_(values, base)
                self._ticks_vec.append(disc_bucks)
                self.shape.append(len(disc_bucks))
                if len(disc_bucks) > max_disc_bins:
                    max_disc_bins = len(disc_bucks)
            else:
                min_lgpv = np.log(values.min()) * 1.0 / np.log(base)
                max_lgpv = np.log(values.max()) * 1.0 / np.log(base)
                N_bins = int(np.ceil(max_disc_bins / 10.0)) * 10
                bucks = min_lgpv + (max_lgpv - min_lgpv) * np.arange(N_bins + 1) * 1.0 / N_bins
                self._ticks_vec.append(np.power(base, bucks))
                self.shape.append(len(bucks))

    def histogram_gen(self, method="degree", **argv):
        """
        histogram generator with heuristic strategies specified with parameter [method]
        :param method: specify the histogram strategy:  "degree" | "mdisc" | "N" | "logP"
        :param argv:  variable argument for specified strategy.
                method   |   argv
                degree   |   base (default=10)
                mdisc    |   discindex   a list indicating the discrete feature index
                         |   base  (default=10)
                N        |   N (or N(s) a list for each dimension feature)
                         |   logarithmic (default=False), 
                         |   base(default=10, if logarithmic is True)
                logP     |   P (the log-base) (or P(s) a list for each dimension feature)
        :return:
        """
        heuristic = ["degree", "N", "logP"]
        if method not in heuristic:
            ValueError("Parameter Error: heuristic strategy should be in {}".format(heuristic))
            exit()

        self.shape = list()
        self._ticks_vec = list()
        if method is "degree":
            # N = 100
            base = 10
            if argv.get("N", None) is not None:
                N = 100
            if argv.get("base", None) is not None:
                base = argv["base"]
            self._check_data_()
            self._degree_feature_bucketize_(base)
        elif method is 'mdisc':
            base = 10
            mdisc_index = None
            if argv.get("mdisc_index", None) is not None:
                mdisc_index = argv["mdisc_index"]
            if argv.get("base", None) is not None:
                base = argv["base"]
            self._check_data_()
            self._multidiscrete_bucketize_(mdisc_index, base)
        elif method is "N":
            N, logarithmic, base = argv.get("N"), argv.get("logarithmic", False), argv.get("base", 10)
            Ns = list()
            if isinstance(N, int):   Ns = list([N]) * self._mode
            elif isinstance(N, list):  Ns = list(N)
            else:
                print("Parameter Error: N should be an integer or a list for bucketize the features")
                exit()

            if logarithmic is True:
                self._check_data_()

            for mod in range(self._mode):
                bucks = self._bucketize_N_(self._features[:, mod], Ns[mod], logarithmic, base)
                self._ticks_vec.append(bucks)
                self.shape.append(len(bucks))
        else:
            self._check_data_()
            P = argv.get("P")
            Ps = list()
            if isinstance(P, int):  Ps = list([P]) * self._mode
            elif isinstance(P, list):  Ps = list(P)
            else:
                print("Parameter Error: P should be an integer or a list for bucketize the features")
                exit()

            for mod in range(self._mode):
                bucks = self._bucketize_logP_(self._features[:, mod], Ps[mod])
                self._ticks_vec.append(bucks)
                self.shape.append(len(bucks))

        points_coord = list()
        for mod in range(self._mode):
            _, pts_pos, _ = self._seq2hists_(self._features[:, mod], self._ticks_vec[mod])
            pts_pos[pts_pos < 0] = 0
            points_coord.append(pts_pos)

        self.points_coord = np.asarray(points_coord, int).T
        self.histogram = dict()
        for pos in self.points_coord:
            self.histogram[tuple(pos)] = self.histogram.get(tuple(pos), 0) + 1

        self.hpos2avgfeat = dict()
        for i in range(self._npts):
            pos, feat = tuple(self.points_coord[i]), self._features[i]
            self.hpos2avgfeat[pos] = self.hpos2avgfeat.get(pos, np.zeros_like(feat)) + feat

        for pos in self.hpos2avgfeat.keys():
            self.hpos2avgfeat[pos] /= 1.0 * self.histogram[pos]

    def save_histogram(self, outfn, sep=',', comments="#"):
        with open(outfn, 'w') as ofp:
            header = comments + " {}".format(str(self.shape)[1:-1].replace(', ', sep))
            ofp.writelines(header + "\n")
            for vec in self._ticks_vec:
                ticks_str = sep.join(map(str, vec))
                ofp.writelines("# " + ticks_str + "\n")

            for pos, cnt in self.histogram.items():
                pos_str = sep.join(map(str, pos))
                line = "{}{}{}".format(pos_str, sep, cnt)
                ofp.writelines(line + "\n")
            ofp.close()

    def save_pts_index(self, outfn, sep=',', comments="#", pts_idx=None):
        header = comments + "pts: {}; ranges: {}".format(len(self._features), str(self.shape)[1:-1].replace(', ', sep))
        index = np.arange(self._npts) if pts_idx is None else np.asarray(pts_idx, int)
        out_dt = np.vstack([index, self.points_coord.T]).T
        np.savetxt(outfn, out_dt, '%d', sep, header=header)
        # print("save points2pos done!")

    def save_hpos2avgfeat(self, outfn, sep=',', comments="#"):
        with open(outfn, 'w') as ofp:
            pos_xs, feat_vs = "", ""
            for k in range(self._mode):
                pos_xs += "hpos-{}{}".format(k+1, sep)
                feat_vs += "avgfeat-{}{}".format(k+1, sep)
            headers = pos_xs + feat_vs + "#npts"
            ofp.writelines(comments + " " + headers + '\n')
            for pos, feat in self.hpos2avgfeat.items():
                pos_str = sep.join(map(str, pos))
                feat_str = str(feat.tolist())[1:-1].replace(', ', sep)
                line = pos_str + sep + feat_str + sep + str(self.histogram[pos])
                ofp.writelines(line + '\n')
            ofp.close()

    def dump(self):
        print("Histogram Info:")
        print("\t Histogram shape: {}".format(self.shape))
        print("\t #points: {}, #mode: {}".format(self._npts, self._mode,))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   graph.py
@Desc    :   Definition of graph structure.
'''

# here put the import lib
from . import STensor
import spartan as st
import numpy as np


class Graph:
    def __init__(self, graph_tensor: STensor, weighted: bool = False,
                 bipartite: bool = False, modet=None):
        '''Construct a graph from sparse tensor.
        If the sparse tensor has more than 2 modes, then it is a rich graph.
        Parameters:
        ------
        modet: int
            The order of mode in graph tensor for temporal bins if exit, start from zero.
            Default is 3.
        '''
        self.graph_tensor = graph_tensor
        self.weighted = weighted
        self.bipartite = bipartite
        self.modet = modet  # which mode is time dimension
        self.nprop = graph_tensor.ndim - 2  # num of edge properties

        self.sm = graph_tensor.sum_to_scipy_sparse(modes=(0, 1))
        if not weighted:
            self.sm = (self.sm > 0).astype(int)
        if not bipartite:
            self.sm = self.sm.maximum(self.sm.T)

    def get_time_tensor(self):
        '''Get the tensor only have time dimension.
        If nprop == 1 and modet == 3, then the tensor is graph_tensor itself.
        If modet is None, then None is returned.
        '''
        return self.get_one_prop_tensor(self.modet)

    def get_one_prop_tensor(self, mode):
        '''Get the tensor only have one edge-property dimension.
        if nprop == 1 and mode == 3, then the tensor is graph_tensor itself.
        If mode is None, and other invalidation, then None is returned.
        '''
        if self.nprop == 1 and mode == 3:
            return graph_tensor
        elif self.nprop > 1 and mode is not None and\
                mode < self.nprop + 2:
            return STensor((self.graph_tensor.coords[(0, 1, mode), :],
                            self.graph_tensor.data))
        else:
            return None

    def get_sub_graph(self, rows, cols):

        cootensor = self.graph_tensor

        gr = -1 * st.ones(cootensor.shape[0], dtype=int)
        gc = -1 * st.ones(cootensor.shape[1], dtype=int)

        lr = len(rows)
        lc = len(cols)

        ar = st.arange(0, lr, 1)
        ac = st.arange(0, lc, 1)
        gr[rows[ar]] = ar
        gc[cols[ac]] = ac
        mrow = cootensor.coords[0]
        mcol = cootensor.coords[1]
        newelem = (gr[mrow] > -1) & (gc[mcol] > -1)

        newrows = mrow[newelem]
        newcols = mcol[newelem]

        subcoords = st.stack((gr[newrows], gc[newcols],
            *cootensor.coords[2:,newelem]), axis=0)
        subvalues = cootensor.data[newelem]

        subtensor = st.STensor((subcoords, subvalues),
                shape=(lr,lc,*cootensor.shape[2:]) )

        return st.Graph(subtensor, self.weighted, self.bipartite, self.modet)

    def get_subgraph_nedges(self, rows, cols):
        """
        Pulls out an arbitrary i.e. non-contiguous submatrix out of
        a sparse.coo_matrix.

        Returns
        ------
        tuples of org_row_id, org_col_id, value
        """
        matr = self.sm.tocoo()

        gr = -1 * st.ones(matr.shape[0], dtype=int)
        gc = -1 * st.ones(matr.shape[1], dtype=int)

        lr = len(rows)
        lc = len(cols)

        ar = st.arange(0, lr, 1)
        ac = st.arange(0, lc, 1)
        gr[rows[ar]] = ar
        gc[cols[ac]] = ac
        mrow = matr.row
        mcol = matr.col
        newelem = (gr[mrow] > -1) & (gc[mcol] > -1)
        subvalues = matr.data[newelem]

        if self.weighted:
            nedges = len(subvalues)
        else:
            nedges = subvalues.sum()

        return nedges

    def degrees(self):
        rowdegs, coldegs = self.sm.sum(axis=1), self.sm.sum(axis=0)
        return rowdegs.A1, coldegs.A1

    def get_edgelist_array(self):
        """
        Get the edgelist of graph (without edge attributes),
        summing up weights of the same edge.

        Returns
        ------
        2D numpy ndarray of (row, col, weights) as list.
        """
        coosm = self.sm.tocoo()
        data = coosm.data
        row = coosm.row
        col = coosm.col
        return np.vstack((row, col, data)).T





# A tree data structure which stores a list of degrees and can quickly retrieve the min degree element,
# or modify any of the degrees, each in logarithmic time. It works by creating a binary tree with the
# given elements in the leaves, where each internal node stores the min of its two children.
import math
class MinTree:
    def __init__(self, degrees):
        self.height = int(math.ceil(math.log(len(degrees), 2)))
        self.numLeaves = 2 ** self.height
        self.numBranches = self.numLeaves - 1
        self.n = self.numBranches + self.numLeaves
        self.nodes = [float('inf')] * self.n
        for i in range(len(degrees)):
            self.nodes[self.numBranches + i] = degrees[i]
        for i in reversed(range(self.numBranches)):
            self.nodes[i] = min(self.nodes[2 * i + 1], self.nodes[2 * i + 2])

    # @profile
    def getMin(self):
        cur = 0
        for i in range(self.height):
            cur = (2 * cur + 1) if self.nodes[2 * cur + 1] <= self.nodes[2 * cur + 2] else (2 * cur + 2)
        # print("found min at %d: %d" % (cur, self.nodes[cur]))
        return (cur - self.numBranches, self.nodes[cur])

    # @profile
    def changeVal(self, idx, delta):
        cur = self.numBranches + idx
        self.nodes[cur] += delta
        for i in range(self.height):
            cur = (cur - 1) // 2
            nextParent = min(self.nodes[2 * cur + 1], self.nodes[2 * cur + 2])
            if self.nodes[cur] == nextParent:
                break
            self.nodes[cur] = nextParent

    def getVal(self, idx):
        cur = self.numBranches + idx
        return self.nodes[cur]

    def setVal(self, idx, val):
        cur = self.numBranches + idx
        self.nodes[cur] = val
        for i in range(self.height):
            cur = (cur - 1) // 2
            nextParent = min(self.nodes[2 * cur + 1], self.nodes[2 * cur + 2])
            if self.nodes[cur] == nextParent:
                break
            self.nodes[cur] = nextParent

    def dump(self):
        print("numLeaves: %d, numBranches: %d, n: %d, nodes: " % (self.numLeaves, self.numBranches, self.n))
        cur = 0
        for i in range(self.height + 1):
            for j in range(2 ** i):
                print(self.nodes[cur])
                cur += 1
            print()



import math
import warnings
from collections import Counter
from itertools import combinations

import numpy as np
import scipy.sparse as ssp

import spartan as st
from .._model import DMmodel
from spartan import STensor


class kGS(DMmodel):
    def __init__(self, graph) -> None:
        self.sm = graph.sm.tolil()
        self.sm.setdiag(0)
        self.N = self.sm.shape[0]
        self.nodes = list(range(self.N))

        self.adj = [Counter(self.sm.rows[n]) for n in range(self.N)]
        self.sizes = [1] * self.N
        self.nodes_dict = dict(
            zip(range(self.N), ({n} for n in range(self.N))))

    def _greedy(self):
        min_pair, min_loss = None, math.inf
        for u, v in combinations(self.nodes):
            loss = self._merge_loss(u, v)
            if loss < min_loss:
                min_loss = loss
                min_pair = (u, v)
        return min_pair, min_loss

    def _sample_pairs(self, C=0):
        if callable(C):
            C = C()
        elif C == 0:
            C = int(math.log2(self.numNode))
        C = int(C)

        min_pair, min_loss = None, math.inf
        for _ in range(C):
            u, v = np.random.choice(self.nodes, 2)
            loss = self._merge_loss(u, v)
            if loss < min_loss:
                min_loss = loss
                min_pair = (u, v)
        return min_pair, min_loss

    def _linear_check(self):
        u = np.random.choice(self.nodes, 1)[0]

        min_pair, min_loss = None, math.inf
        for v in self.nodes:
            if v == u:
                continue
            loss = self._merge_loss()
            if loss < min_loss:
                min_loss = loss
                min_pair = (u, v)
        return min_pair, min_loss

    def run(self, K, strategy='sample_pairs'):
        return self._summarize(K, strategy)

    def summarization(self, K, strategy='sample_pairs'):
        return self._summarize(K, strategy)

    def _check_parameters(self, K, strategy):
        if K >= self.N:
            print(f"`K`({K}) should be less than size of graph({self.N})")
            return False
        if strategy not in ('greedy', 'linear_check', 'sample_pairs'):
            print(f"`Strategy`({strategy}) must be of: 'greedy', 'linear_check', 'sample_pairs'")
            return False
        if strategy == 'greedy' and self.N >= 10000:
            warnings.warn("Using greedy strategy on large graphs is time-consuming, try using other strategies instead.")
        return True


    def _summarize(self, K, strategy='sample_pairs'):
        if not self._check_parameters(K, strategy):
            print("Check parameter fails")
            return
        self.numNode = self.N
        while self.numNode > K:
            pair = None
            if strategy == 'greedy':
                pair, loss = self._greedy()
            elif strategy == 'linear_check':
                pair, loss = self._linear_check()
            elif strategy == 'sample_pairs':
                pair, loss = self._sample_pairs()
            if pair is None:
                break
            u, v = pair
            self._merge(u, v)

        l1_error = self._final_errors()
        print(f"Summarize from {self.N} nodes to {self.numNode} nodes.")
        print(f"Average L1 loss: {l1_error / (self.N * self.N):.4f}")

        rows, cols, datas = [], [], []
        for i, (n, nodes) in enumerate(self.nodes_dict.items()):
            datas.extend([1.0 / len(nodes)] * len(nodes))

        rows, cols, datas = [], [], []
        for i, (n, nodes) in enumerate(self.nodes_dict.items()):
            assert len(nodes) != 0
            nodes = list(nodes)
            rows.extend([i] * len(nodes))
            cols.extend(nodes)
            datas.extend([1.0 / len(nodes)] * len(nodes))

        P_ = ssp.csr_matrix(([1] * len(datas), (rows, cols)), shape=(len(self.nodes_dict), self.N))
        sm_s = P_ @ self.sm @ P_.T

        self.nodes_dict = self.nodes_dict
        self.sm_s = sm_s

        return STensor.from_scipy_sparse(sm_s)

    def _merge_loss(self, u, v):
        sizeu = self.sizes[u]
        sizev = self.sizes[v]
        if sizeu == 0 or sizev == 0:
            return math.inf

        loss = 0.0
        new_e = 0
        if u in self.adj[u]:
            eu = self.adj[u][u]
            new_e += eu
            loss -= 8.0 * eu * eu / (sizeu * (sizeu-1))
        if v in self.adj[v]:
            ev = self.adj[v][v]
            new_e += ev
            loss -= 8.0 * ev * ev / (sizev * (sizev-1))

        for n in self.adj[u]:
            value = self.adj[u][n]
            if n == u or value == 0 or self.sizes[n] == 0:
                continue
            loss -= 4.0 * value * value / (self.sizes[u] * self.sizes[n])
        for n in self.adj[v]:
            value = self.adj[v][n]
            if n == u or value == 0 or self.sizes[n] == 0:
                continue
            loss -= 4.0 * value * value / (self.sizes[v] * self.sizes[n])

        if v in self.adj[u]:
            new_e += 2 * self.adj[u][v]
        loss += 8 * new_e * new_e / ((sizeu+sizev) * (sizeu + sizev-1))

        common_neis = self.adj[u].keys() & self.adj[v].keys()
        for n in common_neis:
            if n == u or n == v or self.sizes[n] == 0:
                continue
            eui = self.adj[u][n]
            evi = self.adj[v][n]
            if eui == 0 and evi == 0:
                continue
            loss += 4.0 / (sizeu + sizev) * (eui * evi + evi *
                                             evi + 2.0 * eui * evi) / self.sizes[n]

        return loss

    def _merge(self, u, v):
        nodes_dict = self.nodes_dict

        nodes_dict[u] = nodes_dict[u] | nodes_dict[v]
        del nodes_dict[v]
        self.sizes[u] += self.sizes[v]
        self.sizes[v] = 0

        # Update adj list
        uv = self.adj[u][v]
        self.adj[u].update(self.adj[v])
        self.adj[u][u] += (uv + self.adj[v][v])
        if self.adj[u][u] == 0:
            del self.adj[u][u]
        del self.adj[u][v]
        self.adj[v] = None
        for nei in self.adj[u]:
            if nei != u and nei != v and self.sizes[nei] != 0:
                if self.adj[u][nei] != 0:
                    self.adj[nei][u] = self.adj[u][nei]

        self.nodes.remove(v)
        self.numNode -= 1

    def _final_errors(self):
        l1_error = 0.0
        for n in self.nodes:
            sizesn = self.sizes[n]
            if sizesn == 0:
                continue
            for nei in self.adj[n]:
                sizes_nei = self.sizes[nei]
                if sizes_nei == 0:
                    continue
                if nei == n:
                    w = self.adj[n][n] / (sizesn * (sizesn - 1))
                    l1_error += (1-w)
                else:
                    w = self.adj[n][nei] / (sizesn * sizes_nei)
                    l1_error += (1-w)

        l1_error *= 2
        for n in self.nodes:
            for nei in self.adj[n]:
                if self.sizes[nei] == 0:
                    continue
                l1_error += self.adj[n][nei]

        return l1_error


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py
@Desc    :   Import part and configuration part, including names exposed to spartan.
'''

# here put the import lib

from spartan.util import MODEL_PATH

from enum import Enum
from functools import partial

from ._model import PipeLine


def __call__(policy: str, *args, **kwargs) -> object:
    '''Design for dynamic import of call by model.

    Use partial function to point out specific path of each model.

    Parameters
    ----------
    policy : str
        path of model

    Returns
    ----------
    model_obj
        object for model
    '''
    import importlib
    model_cls = importlib.import_module(policy).__call__()
    return model_cls.__create__(*args, **kwargs)


# please register a new module in lexical (alpha-beta) order
BeatLex = partial(__call__, MODEL_PATH + ".beatlex")
BeatGAN = partial(__call__, MODEL_PATH + ".beatgan")
EagleMine = partial(__call__, MODEL_PATH + ".eaglemine")
EigenPulse = partial(__call__, MODEL_PATH + ".eigenpulse")
Eigenspokes = partial(__call__, MODEL_PATH + ".eigenspokes")
HoloScope = partial(__call__, MODEL_PATH + ".holoscope")
FlowScope = partial(__call__, MODEL_PATH + ".flowscope")
RPeak = partial(__call__, MODEL_PATH + ".rpeak")
DPGS = partial(__call__, MODEL_PATH + ".DPGS")
kGrass = partial(__call__, MODEL_PATH + ".kGS")
IAT = partial(__call__, MODEL_PATH + ".iat")
Fraudar = partial(__call__, MODEL_PATH + ".fraudar")
CubeFlow = partial(__call__, MODEL_PATH + ".CubeFlow")
Specgreedy = partial(__call__, MODEL_PATH + ".specgreedy")
# TODO Specgreedy = partial(__call__, MODEL_PATH + ".specgreedy.Specgreedy.Specgreedy")
CubeFlowPlus = partial(__call__, MODEL_PATH + ".CubeFlowPlus")


__all__ = [
    'PipeLine',
    'BeatLex',
    'BeatGAN',
    'EagleMine',
    'EigenPulse',
    'Eigenspokes',
    'HoloScope',
    'FlowScope',
    'RPeak',
    'DPGS',
    "kGrass",
    'IAT',
    'Fraudar',
    'CubeFlow',
    'Specgreedy'
    'CubeFlowPlus'
]



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#  File: views_viz.py
#  Goal: provide visualization tools for viewing the clustering result
#
#      Version:  1.0
#      Goal: Subroutine script
#      Created by @wenchieh  on <12/17/2018>
#

__author__ = 'wenchieh'

# third-party lib
import numpy as np
from scipy.sparse import csr_matrix

# project
from .desc.gaussian_describe import GaussianDescribe
from .desc.dtmnorm_describe import DTMNormDescribe

from .utils.loader import Loader
from .utils.ploter import plot_clusters, plot_heatmap
from .utils.ploter_aux import plot_heatmap_ellipse_covs

VALID_DESCVOCS = ['dtmnorm', 'dmgauss']


def size_relabels(labels):
    clsdic = {}
    for l in labels:
        if l not in clsdic:
            if l != -1: clsdic[l] = len(clsdic)
            else:       clsdic[l] = l
    rlbs = [clsdic[l] for l in labels]
    return np.array(rlbs)

def cluster_view(ins_hcel2lab, outs=None, outlier_lab=-1):
    hcube_label = np.loadtxt(ins_hcel2lab, int, delimiter=',')
    outliers_index = hcube_label[:, -1] == outlier_lab
    outliers = hcube_label[outliers_index, :-1]
    others_cls = hcube_label[~outliers_index]
    labels = others_cls[:, -1]  #size_relabels(others_cls[:, -1])
    cls_fig = plot_clusters(others_cls[:, :-1], [], labels, outliers=outliers[::-1], ticks=False)
    if outs is not None:
        cls_fig.savefig(outs)
    cls_fig.show()

def describe_view(ins_hist, ins_desc, desc_voc, xlab, ylab, outs):
    assert desc_voc in VALID_DESCVOCS
    loader = Loader()
    desc = DTMNormDescribe if desc_voc == 'dtmnorm' else GaussianDescribe
    desc_parms = loader.load_describes_parms(ins_desc, desc, mode=2)
    h_shape, ticks_vec, hist_arr = loader.load_histogram(ins_hist)
    csr_mat = csr_matrix((hist_arr[:, -1], (hist_arr[:, 0], hist_arr[:, 1])), shape=h_shape, dtype=int)
    # plot_heatmap(ticks_vec[1], ticks_vec[0], csr_mat.toarray(), xlabel=xlabel, ylabel=ylabel, outfn=outfn)
    plot_heatmap_ellipse_covs(ticks_vec[1], ticks_vec[0], csr_mat.toarray(), desc_parms, base=10,
                              scales=(1.5, 3), xlabel=xlab, ylabel=ylab, outfn=outs)


if __name__ == '__main__':
    path = '../output/'
    ins_hist = 'histogram.out'
    ins_desc = 'describe.out'
    ins_hcel2lab = 'hcel2label.out'
    desc_voc = 'dtmnorm'
    xlab, ylab = ["Hubness", "Out-degree"]   # ["Authority", "In-degree"]
    viz_clsv = 'viz_cluster.png'
    viz_desc = 'viz_describes.png'

    cluster_view(path + ins_hcel2lab, path + viz_clsv)
    describe_view(path + ins_hist, path + ins_desc, desc_voc, xlab, ylab, path + viz_desc)



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  Describe hypercubes (in high dimensional) with DTM Gaussian distribution
#      DTM Gaussian: Discrete Truncate Multivariate Normal Gaussian.
#  Author: wenchieh
#
#  Project: eaglemine
#      dtmnorm.py
#      Version:  1.0
#      Date: December 1 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/01/2017>
#
__author__ = 'wenchieh'

# sys
import collections

# third-party lib
import numpy as np
from numpy import inf
from numpy.linalg import det
import scipy.optimize as opt
from statsmodels.stats.weightstats import DescrStatsW
from sklearn import cluster
from pomegranate import GeneralMixtureModel, MultivariateGaussianDistribution

# project
from .norm_extras import mvnormcdf


class DTMNorm(object):
    def __init__(self, ndim=2, lower_bound=None, upper_bound=None):
        self.ndim = ndim
        if lower_bound is None:
            lower_bound = np.array([0] * ndim)
        self.lowers = np.array(lower_bound)
        if upper_bound is None:
            upper_bound = np.array([inf] * ndim)
        self.uppers = np.array(upper_bound)

    def set_lowers(self, lower_bound):
        self.lowers = np.array(lower_bound)

    def get_lowers(self):
        return self.lowers

    def set_uppers(self, upper_bound):
        self.uppers = np.array(upper_bound)

    def get_uppers(self):
        return self.uppers

    def _paras_compose_(self, mus, covs, ws):
        paras = list()
        paras.extend(list(np.array(mus).flatten()))

        for i in range(len(covs)):
            cov = covs[i]
            for k in range(self.ndim):
                paras.extend(np.array(cov[k, k:]).tolist())

        if ws is not None:
            paras.extend(ws[:-1])
        return np.array(paras)

    def _paras_decompose_(self, paras, n_components):
        mus, covs, ws = list(), list(), list()
        mus = list(np.array(paras[: self.ndim * n_components]).reshape((n_components, self.ndim)))
        covupt_len = range(self.ndim, 0, -1)
        for i in range(n_components):
            cov_upt = np.zeros((self.ndim, self.ndim))
            covs_starts = self.ndim * n_components + i * np.sum(covupt_len)
            for k in range(self.ndim):
                start = int(covs_starts + np.sum(covupt_len[:k]))
                end = int(covs_starts + np.sum(covupt_len[:k+1]))
                cov_upt[k, k:] = np.array(paras[start: end])
            cov = cov_upt + cov_upt.T - np.diag(np.diag(cov_upt))
            covs.append(cov)

        mc_parm_len = len(mus) + n_components * np.sum(covupt_len)
        if mc_parm_len < len(paras):
            ws = list(paras[-(n_components-1):])
            ws = ws + list([1.0 - np.sum(ws)])
        else:
            ws = [1.0 / n_components] * n_components
        return mus, covs, np.asarray(ws)

    #TODO: non-positive covariance matrix process [refinement]
    def _cov_process_(self, cov, reg_covar=1e-4):
        '''
        process the un-expected case: cov is non-positive definitive covariance
            only use diagonal matrix of cov to replace current cov.
        other suggestions:
               reference to:
                   Not Positive Definite Matrices--Causes and Cures
                   [http://www2.gsu.edu/~mkteer/npdmatri.html]
        '''
        ## Non-negative regularization added to the diagonal of covariance.
        ## Allows to assure that the covariance matrices are all positive.
        if det(cov) <= 0.0:
            diags = np.diag(cov)
            if 0.0 in diags:
                diags += reg_covar
            cov = np.diag(diags)
            # diag_min = np.min(diags)
            # if diag_min < 0:
            #     for k in range(self.ndim):
            #         cov[k, k] -= diag_min
        return cov

    def _single_optpara(self, paras_vec, pos_left, pos_right, weights, debug_info=None):
        res = 0.0
        N = len(pos_left)
        mus, covs, _ = self._paras_decompose_(paras_vec, 1)
        cov = self._cov_process_(covs[0])

        # smoothed normalizer
        normalizer = np.max([8e-3, mvnormcdf(self.uppers, mus[0], cov, self.lowers)])
        for i in range(N):
            numerator = np.max([0, mvnormcdf(pos_right[i, :], mus[0], cov, pos_left[i, :])])
            P_pos = numerator / normalizer
            if P_pos > 0:
                res += weights[i] * np.log(P_pos)
            else:
                res += -1.0 * np.NaN

        if debug_info is not None:
            debug_info.append(-res)
            if len(debug_info) % 200  == 0:
                print(len(debug_info), -res)

        return -1 * res

    def fit_single(self, pos_left, pos_right, weights, tol=1e-4, maxiter=4000, verbose=False):
        left, right = np.asarray(pos_left), np.asarray(pos_right)
        debugs = list() if verbose else None
        centers = (left + right) / 2.0
        statsW = DescrStatsW(centers, weights=np.array(weights))
        init_paras = self._paras_compose_([statsW.mean], [statsW.cov], [1.0])

        method = 'Nelder-Mead'
        res = opt.minimize(self._single_optpara, init_paras, args=(left, right, weights, debugs),
                           method = method, tol = tol, options={'maxiter': maxiter, 'disp': verbose})
        if verbose:
            print("Method:{}; Initial parameter: {};".format(method, init_paras))
            print("Converged Parameter: {}".format(res.x))

        mus, covs, ws = self._paras_decompose_(res.x, 1)
        if det(covs[0]) == 0.0:
            print("Warning: covariance processed:")
            print("\t pre-optimal mus: {}, cov: {}".format(mus[0], covs[0]))
            covs[0] = self._cov_process_(covs[0])

        return mus, covs, ws, res.fun

    def _mixture_optpara(self, paras_vec, pos_left, pos_right, weights, n_components=2, debug_info=None):
        res = 0.0
        N = len(pos_left)
        mus, covs, comp_ws = self._paras_decompose_(paras_vec, n_components)
        normalizors = list()
        for k in range(n_components):
            covs[k] = self._cov_process_(covs[k])
            denorm = np.max([8e-3, mvnormcdf(self.uppers, mus[k], covs[k], self.lowers)])
            normalizors.append(denorm)

        for i in range(N):
            ps_pos = [np.max([1e-10, mvnormcdf(pos_right[i, :], mus[k], covs[k], pos_left[i, :])])
                      for k in range(n_components)]
            prob = np.sum(ps_pos * comp_ws)
            if prob > 0:
                res += weights[i] * np.log(prob)
            else:
                res += -1.0 * np.NaN

        if debug_info is not None:
            debug_info.append(-res)
            if len(debug_info) % 200  == 0:
                print(len(debug_info), -res)

        return -1 * res

    def fit_mixture(self, pos_left, pos_right, weights, n_components=2, tol=1e-4, maxiter=4000, verbose=False):
        left, right = np.asarray(pos_left), np.asarray(pos_right)
        weights = np.asarray(weights)
        debugs = list() if verbose else None
        centers = (left + right) / 2.0
        init_gmm = GeneralMixtureModel.from_samples(MultivariateGaussianDistribution,
                                                   n_components=n_components, X=centers, weights=weights,
                                                   stop_threshold=0.01, n_jobs=2)

        init_mus, init_covs = list(), list()
        init_comp_ws = np.array(init_gmm.weights)
        init_comp_ws /= np.sum(init_comp_ws)
        for i in range(n_components):
            paras = init_gmm.distributions[i].parameters
            init_mus.append(np.array(paras[0]))
            init_covs.append(np.array(paras[1]))

        init_paras = self._paras_compose_(init_mus, init_covs, list(init_comp_ws))

        method = 'Nelder-Mead'
        res = opt.minimize(self._mixture_optpara, init_paras, args=(left, right, weights, n_components, debugs),
                           method = method, tol = tol, options={'maxiter': maxiter, 'disp': verbose})
        if verbose:
            print("Method:{}; Initial parameter: {};".format(method, init_paras))
            print("Converged Parameter: {}".format(res.x))

        mus, covs, comp_ws = self._paras_decompose_(res.x, n_components)
        return mus, covs, comp_ws, res.fun

    def log_loss(self, paras, pos_left, pos_right, weights, is_mix, n_components=2):
        loss = 0.0
        paras_vec = self._paras_compose_(paras['mus'], paras['covs'], paras['weights'])
        if is_mix:
            loss = self._mixture_optpara(paras_vec, pos_left, pos_right, weights, n_components)
        else:
            loss = self._single_optpara(paras_vec, pos_left, pos_right, weights)

        return loss



import builtins
import functools

import numpy as np
import sparse

from .tensor import DTensor, STensor, _ensure_tensor, _wrap_ret


def _dispatch(dfunc, sfunc=None):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            is_sparse = builtins.any([isinstance(x, STensor) for x in args])
            is_dense = builtins.any([isinstance(x, DTensor) for x in args])
            if (is_sparse and is_dense):
                raise TypeError(
                    f"Parameters of `st.{func.__name__}` should be all STensor \
                    or all DTensor.")
            if is_sparse:
                if sfunc is None:
                    raise TypeError(f"`st.{func.__name__}` doesn't support \
                                    STensor parameters.")
                args = tuple(x._data if isinstance(x, STensor) else x
                             for x in args)
                return _ensure_tensor(sfunc(*args, **kwargs))
            else:
                args = tuple(x._data if isinstance(x, DTensor) else x
                             for x in args)
                return _ensure_tensor(dfunc(*args, **kwargs))
        return wrapper
    return decorator


# Type definition
short = np.short
uint8 = np.uint8
int8 = np.int8
int16 = np.int16
int32 = np.int32
int64 = np.int64
float16 = np.float16
float32 = np.float32
float64 = np.float64
complex64 = np.complex64
complex128 = np.complex128


@_wrap_ret
@_dispatch(np.add, np.add)
def add(input_, other):
    """Wrapper of `numpy.add`.

    Parameters
    ----------
    input_ : DTensor or STensor
        The first operand.
    other : DTensor or STensor
        The second operand.

    Returns
    -------
    DTensor or STensor
        Output tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.all, sparse.COO.all)
def all(input_, axis=None, keepdims=False):
    """Wrapper of `numpy.all` and `sparse.COO.all`

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False

    Returns
    -------
    DTensor or STensor or bool
        Output tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.angle)
def angle(input_, deg=False):
    """Wrapper of `numpy.angle`

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    deg : bool, optional
        If true, result is in degree format. Otherwise, return in radians. By
        default False
    """
    pass


@_wrap_ret
@_dispatch(np.any, sparse.COO.any)
def any(input_, axis=None, keepdims=False):
    """Wrapper of `numpy.any` and `sparse.COO.any`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False

    Returns
    -------
    DTensor or STensor or bool:
    """
    pass


@_wrap_ret
def arange(start, stop, step, dtype=None):
    """Wrapper of `numpy.arange`

    Parameters
    ----------
    start : number
        Start of the interval.
    stop : number
        End of the interval.
    step : number
        Spacing between spaces.
    dtype : data-type, optional
        Type of the return tensor, by default None

    Returns
    -------
    DTensor
        Constructed dense tensor.
    """
    return np.arange(start, stop, step, dtype=dtype)


@_wrap_ret
@_dispatch(np.argmax)
def argmax(input_, axis=None):
    """Wrapper of `numpy.argmax`

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.argmin)
def argmin(input_, axis=None):
    """Wrapper of `numpy.argmin`

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.argsort)
def argsort(input_, axis=-1):
    """Wrapper of `numpy.argsort`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.bincount)
def bincount(input_, weights=None, minlength=0):
    """Wrapper of `numpy.bincount`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    weights : DTensor
        Weights.
    minlength : int, optional
        Minimum number of bins, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.bitwise_and)
def bitwise_and(input_, other):
    """Wrapper of `numpy.bitwise_and`

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.bitwise_not)
def bitwise_not(input_):
    """Wrapper of `numpy.bitwise_not`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.bitwise_or)
def bitwise_or(input_, other):
    """Wrapper of `numpy.bitwise_or`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.bitwise_xor)
def bitwise_xor(input_, other):
    """Wrapper of `numpy.bitwise_xor`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_dispatch(np.can_cast)
def can_cast(from_, to):
    """Wrapper of `numpy.can_cast`.

    Parameters
    ----------
    from_ : data-type
        Data type to cast from.
    to : data-type
        Data type to cast to.
    """
    pass


@_wrap_ret
@_dispatch(np.ceil)
def ceil(input_):
    """Wrapper of `numpy.ceil`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.concatenate, sparse.concatenate)
def concatenate(inputs, axis=None):
    """Wrapper of `numpy.concatenate` and `sparse.concatenate`

    Parameters
    ----------
    inputs : sequence of STensor/DTensor
        Tensor to be concatenated.
    axis : int or None, optional
        Axis to operate on, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.conj, sparse.COO.conj)
def conj(input_):
    """Wrapper of `numpy.conj` and `sparse.COO.conj`.

    Parameters
    ----------
    input_ : DTensor or STensor.
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.cos)
def cos(input_):
    """Wrapper of `numpy.cos`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.cosh)
def cosh(input_):
    """Wrapper of `numpy.cosh`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.cumprod)
def cumprod(input_, axis=None, dtype=None):
    """Wrapper of `numpy.cumprod`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int, optional
        Axis to operate on, by default None
    dtype : data-type, optional
        Data type of output, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.cumsum)
def cumsum(input_, axis=None, dtype=None):
    """Wrapper of `numpy.cumsum`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int, optional
        Axis to operate on, by default None
    dtype : data-type, optional
        Data type of output, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.diag)
def diag(input_, k=0):
    """Wrapper of `numpy.diag`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    k : int, optional
        Offset to main diagonal, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.diagflat)
def diagflat(input_, k=0):
    """Wrapper of `numpy.diagflat`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    offset : int, optional
        Offset to main diagonal, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.diagonal, sparse.diagonal)
def diagonal(input_, offset=0, axis1=0, axis2=1):
    """Wrapper of `numpy.diagonal`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    offset : int, optional
        Offset to the main diagonal, by default 0
    axis1 : int, optional
        The first axis of the diagonal, by default 0
    axis2 : int, optional
        The second axis of the diagonal, by default 1
    """
    pass


@_wrap_ret
@_dispatch(np.dot, sparse.dot)
def dot(input_, other):
    """Wrapper of `numpy.dot` and `sparse.dot`.

    Parameters
    ----------
    input_ : DTensor or STensor
        The first operand.
    other : DTensor or STensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.empty)
def empty(shape, dtype):
    """Wrapper of `numpy.empty`.

    Parameters
    ----------
    shape : tuple of ints
        Shape of tensor.
    dtype : data-type
        Data type of tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.empty_like)
def empty_like(input_, dtype):
    """Wrapper of `numpy.empty_like`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    dtype : data-type
        Data type of output.
    """
    pass


@_wrap_ret
@_dispatch(np.equal)
def equal(input_, other):
    """Wrapper of `numpy.equal`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.exp)
def exp(input_):
    """Wrapper of `numpy.exp`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.expm1, np.expm1)
def expm1(input_):
    """Wrapper of `numpy.expm1`.

    Parameters
    ----------
    input_ : DTensor or STensor.
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.eye, sparse.eye)
def eye(n, m=None, dtype=None):
    """Wrapper of `numpy.eye`.

    Parameters
    ----------
    n : int
        Number of rows.
    m : int, optional
        Number of cols, by default None (equal to `n`)
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.flip)
def flip(input_, axis=None):
    """Wrapper of `numpy.flip`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.floor)
def floor(input_):
    """Wrapper of `numpy.floor`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.floor_divide)
def floor_divide(input_, other):
    """Wrapper of `numpy.floor_divide`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.fmod)
def fmod(input_, other):
    """Wrapper of `numpy.fmod`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.full, sparse.full)
def full(shape, value, dtype=None):
    """Wrapper of `numpy.full`.

    Parameters
    ----------
    shape : tuple of ints
        Shape of output tensor.
    value : scalar
        Fill value of output tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.full_like, sparse.full_like)
def full_like(input_, value, dtype=None):
    """Wrapper of `numpy.full_like` and `sparse.full_like`.

    Parameters
    ----------
    input_ : DTensor or STensor
        The input tensor.
    value : scalar
        Fill value of output tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.imag)
def imag(input_):
    """Wrapper of `numpy.imag`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.isfinite)
def isfinite(input_):
    """Wrapper of `numpy.isinfinite`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.isinf)
def isinf(input_):
    """Wrapper of `numpy.isinf`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.isnan)
def isnan(input_):
    """Wrapper of `numpy.isnan`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.linspace)
def linspace(start, end, step, dtype=None):
    """Wrapper of `numpy.linspace`.

    Parameters
    ----------
    start : DTensor
        Dense tensor of start points.
    end : DTensor
        Dense tensor of end points.
    step : int
        Number of sequence.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.log)
def log(input_):
    """Wrapper of `numpy.log`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.log10)
def log10(input_):
    """Wrapper of `numpy.log10`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.log1p, np.log1p)
def log1p(input_):
    """Wrapper of `numpy.log1p`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.log2)
def log2(input_):
    """Wrapper of `numpy.log2`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.logical_and)
def logical_and(input_, other):
    """Wrapper of `numpy.logical_and`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.logical_not)
def logical_not(input_):
    """Wrapper of `numpy.logical_not`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.logical_or)
def logical_or(input_, other):
    """Wrapper of `numpy.logical_or`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.logical_xor)
def logical_xor(input_, other):
    """Wrapper of `numpy.logical_xor`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.logspace)
def logspace(start, stop, step, base=10, dtype=None):
    """Wrapper of `numpy.logspace`.

    Parameters
    ----------
    start : DTensor
        Dense tensor of start points.
    stop : DTensor
        Dense tensor of end points.
    step : int
        Number of sequence.
    base : int, optional
        Base of the logarithm, by default 10
    dtype : data-type, optional
        Data type of output, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.matmul, sparse.matmul)
def matmul(input_, other):
    """Wrapper of `numpy.matmul`.

    Parameters
    ----------
    input_ : DTensor or STensor
        The first operand.
    other : DTensor or STensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.mean, sparse.COO.mean)
def mean(input_, axis=None, keepdims=False):
    """Wrapper of `numpy.mean` and `sparse.COO.mean`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    pass


@_wrap_ret
@_dispatch(np.median)
def median(input_, axis=-1, keepdims=False):
    """Wrapper of `numpy.median`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    pass


@_wrap_ret
@_dispatch(np.meshgrid)
def meshgrid(*inputs):
    """Wrapper of `numpy.meshigrid`.
    """
    pass


@_wrap_ret
@_dispatch(np.nonzero, sparse.COO.nonzero)
def nonzero(input_):
    """Wrapper of `numpy.nonzero` and `sparse.COO.nonzero`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.ones, sparse.ones)
def ones(shape, dtype=None):
    """Wrapper of `numpy.ones`.

    Parameters
    ----------
    shape : tuple of ints
        Shape of output tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.ones_like, sparse.ones_like)
def ones_like(input_, dtype=None):
    """Wrapper of `numpy.ones_like` and `sparse.ones_like`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.prod, sparse.COO.prod)
def prod(input_, axis=None, keepdims=False, dtype=None):
    """Wrapper of `numpy.prod` and `sparse.COO.prod`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    dtype : data-type, optional
        Data type of output, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.real)
def real(input_):
    """Wrapper of `numpy.real`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.reciprocal)
def reciprocal(input_):
    """Wrapper of `numpy.reciprocal`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.remainder)
def remainder(input_, other):
    """Wrapper of `numpy.remainder`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.reshape, sparse.COO.reshape)
def reshape(input_, shape):
    """Wrapper of `numpy.rehsape` and `sparse.COO.reshape`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    shape : tuple of ints
        Shape of new tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.roll)
def roll(input_, shift, axis=None):
    """Wrapper of `numpy.roll`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    shift : int or tuple of ints
        Shift numbers of each axis or axes.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.rot90)
def rot90(input_, k=1, axes=(0, 1)):
    """Wrapper of `numpy.rot90`

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    k : int, optional
        Number of rotation times, by default 1
    axes : tuple, optional
        The axes in which the input is rotated, by default (0, 1)
    """
    pass


@_wrap_ret
@_dispatch(np.sign)
def sign(input_):
    """Wrapper of `numpy.sign`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.sin, np.sin)
def sin(input_):
    """Wrapper of `numpy.sin`.

    Parameters
    ----------
    input_ : DTensor or STensor.
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.sinh, np.sinh)
def sinh(input_):
    """Wrapper of `numpy.sinh`.

    Parameters
    ----------
    input_ : DTensor or STensor.
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.split)
def split(input_, indices_or_sections, axis=0):
    """Wrapper of `numpy.split`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    indices_or_sections : int or DTensor
        If integer, specifing part of splits. Else specifing the positions
        where the splits take place.
    axis : int, optional
        The axis to operate on, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.sqrt, np.sqrt)
def sqrt(input_):
    """Wrapper of `numpy.sqrt`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.square, np.square)
def square(input_):
    """Wrapper of `numpy.square`.

    Parameters
    ----------
    input_ : DTensor or STensor.
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.squeeze)
def squeeze(input_, axis=None):
    """Wrapper of `numpy.squeeze`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.stack, sparse.stack)
def stack(inputs, axis=0):
    """Wrapper of `numpy.stack` and `sparse.stack`.

    Parameters
    ----------
    inputs : DTensor and STensor.
        Input tensor.
    axis : int, optional
        Axis to operate on, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.std, sparse.COO.std)
def std(input_, axis=None, keepdims=False):
    """Wrapper of `numpy.std` and `sparse.COO.std`.

    Parameters
    ----------
    input_ : DTensor or STensor.
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    pass


@_wrap_ret
@_dispatch(np.sum, sparse.COO.sum)
def sum(input_, axis=None, dtype=None, keepdims=False):
    """Wrapper of `numpy.sum` and `sparse.COO.sum`.

    Parameters
    ----------
    input_ : DTensor or STensor.
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    dtype : data-type, optional
        Data type of output, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    pass


@_wrap_ret
@_dispatch(np.take)
def take(input_, indices):
    """Wrapper of `numpy.take`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    indices : DTensor
        Indices of extracted values.
    """
    pass


@_wrap_ret
@_dispatch(np.tan, np.tan)
def tan(input_):
    """Wrapper of `numpy.tan`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.tanh, np.tanh)
def tanh(input_):
    """Wrapper of `numpy.tanh`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.tensordot, sparse.tensordot)
def tensordot(input_, other, axes=2):
    """Wrapper of `numpy.tensordot` and `sparse.tensordot`.

    Parameters
    ----------
    input_ : DTensor or STensor
        The first operand.
    other : DTensor or STensor
        The second operand.
    axes : int or tuple of ints, optional
        Axis or axes along which the operation is performed, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.trace, np.trace)
def trace(input_):
    """Wrapper of `numpy.trace`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.transpose, sparse.COO.transpose)
def transpose(input_, axes=None):
    """Wrapper of `numpy.transpose` and `sparse.COO.transpose`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor
    axes : list of ints, optional
       Axes along which the operation is performed, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.tril, sparse.tril)
def tril(input_, k=0):
    """Wrapper of `numpy.tril` and `sparse.tril`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    k : int, optional
        Offset to main diagonal, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.tril_indices)
def tril_indices(n, m=0, k=0):
    """Wrapper of `numpy.tril_indices`.

    Parameters
    ----------
    n : int
        Number of row of output tensor.
    m : int, optional
        Number of column of output tensor, by default 0
    offset : int, optional
        Offset to main diagonal, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.triu, sparse.triu)
def triu(input_, k=0):
    """Wrapper of `numpy.triu` and `sparse.triu`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor
    k : int, optional
        Offset to main diagonal, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.triu_indices)
def triu_indices(n, m=0, k=0):
    """Wrapper of `numpy.triu_indices`.

    Parameters
    ----------
    n : int
        Number of row of output tensor.
    m : int, optional
        Number of column of output tensor, by default 0
    offset : int, optional
        Offset to main diagonal, by default 0
    """
    pass


@_wrap_ret
@_dispatch(np.true_divide)
def true_divide(input_, other):
    """Wrapper of `numpy.true_divide`.

    Parameters
    ----------
    input_ : DTensor
        The first operand.
    other : DTensor
        The second operand.
    """
    pass


@_wrap_ret
@_dispatch(np.trunc)
def trunc(input_):
    """Wrapper of `numpy.trunc`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.unique)
def unique(input_, return_inverse=False, return_counts=False, axis=None):
    """Wrapper of `numpy.unique`.

    Parameters
    ----------
    input_ : DTensor
        Input dense tensor.
    return_inverse : bool, optional
        If True, return the indices also, by default False
    return_counts : bool, optional
        If True, return the count also, by default False
    axis : int or None, optional
        The axis to operate on, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.var, sparse.COO.var)
def var(input_, axis=None, keepdims=False):
    """Wrapper of `numpy.var` and `sparse.COO.var`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    axis : None or int or tuple of ints, optional
        Axis or axes to operate on, by default None
    keepdims : bool, optional
        If true, the axes along which the operation performed are left to size
        one, by default False
    """
    pass


@_wrap_ret
@_dispatch(np.where, sparse.where)
def where(condition, x, y):
    """Wrapper of `numpy.where` and `sparse.where`.

    Parameters
    ----------
    condition : DTensor of bool
        Where True, yield x, otherwise yield y.
    x : DTensor or STensor
        The first tensor.
    y : DTensor or STensor
        The second tensor.
    """
    pass


@_wrap_ret
@_dispatch(np.zeros, sparse.zeros)
def zeros(shape, dtype=None):
    """Wrapper of `numpy.zeros` and `sparse.zeros`.

    Parameters
    ----------
    shape : tuple of ints
        Shape of output tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    pass


@_wrap_ret
@_dispatch(np.zeros_like, sparse.zeros_like)
def zeros_like(input_, dtype=None):
    """Wrapper of `numpy.zeros_like` and `sparse.zeros_like`.

    Parameters
    ----------
    input_ : DTensor or STensor
        Input tensor.
    dtype : data-type, optional
        Data type of output tensor, by default None
    """
    pass
